"""Demonstrate the effect of fixed initial conditions."""

import copy
import numpy as np
import matplotlib.pyplot as plt
from pysb.simulator import ScipyOdeSimulator

from pysb.examples.fixed_initial import model

n_obs = len(model.observables)
tspan = np.linspace(0, 0.5)

plt.figure(figsize=(8,4))

# Simulate the model as written, with free F fixed.
sim = ScipyOdeSimulator(model, tspan)
res = sim.run()
obs = res.observables.view(float).reshape(-1, n_obs)
plt.subplot(121)
plt.plot(res.tout[0], obs)
plt.ylabel('Amount')
plt.xlabel('Time')
plt.legend([x.name for x in model.observables], frameon=False)
plt.title('Free F fixed')

# Make a copy of the model and unfix the initial condition for free F.
model2 = copy.deepcopy(model)
model2.reset_equations()
model2.initials[1].fixed = False
sim2 = ScipyOdeSimulator(model2, tspan)
res2 = sim2.run()
obs2 = res2.observables.view(float).reshape(-1, n_obs)
plt.subplot(122)
plt.plot(res2.tout[0], obs2)
plt.xlabel('Time')
plt.legend([x.name for x in model.observables], frameon=False)
plt.title('Free F can be consumed')

plt.show()



"""Simulate the bax_pore_sequential model and plot the results."""

from __future__ import print_function
import matplotlib.pyplot as plt
from numpy import logspace
from pysb.simulator import ScipyOdeSimulator

from bax_pore_sequential import model, max_size


# System is very stiff, and using logspace instead of linspace to produce the
# vector of time points happens to help with the integration
t = logspace(-3, 5) # 1e-3 to 1e5
print("Simulating...")
x = ScipyOdeSimulator(model).run(tspan=t).all

# Plot trajectory of each pore
for i in range(1, max_size + 1):
    observable = 'Bax%d' % i
    # Map pore size to the central 50% of the YlOrBr color map
    color = plt.cm.YlOrBr(float(i) / max_size / 2 + 0.25)
    plt.plot(t, x[observable], c=color, label=observable)
# Plot Smac species
plt.plot(t, x['mSmac'], c='magenta', label='mSmac')
plt.plot(t, x['cSmac'], c='cyan', label='cSmac')

# Expand the limits a bit to show the min/max levels clearly
plt.ylim([-0.01e5, 1.01e5])
# Show time on a log scale 
plt.xscale('log')
plt.legend(loc='upper right')
plt.show()



"""
Module containing a class for returning the Kappa equivalent for a given PySB model.

Serves as a wrapper around :py:class:`pysb.generator.kappa.KappaGenerator`.

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""

from pysb.generator.kappa import KappaGenerator
from pysb.export import Exporter

class KappaExporter(Exporter):
    """A class for returning the Kappa for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """

    def export(self, dialect='kasim'):
        """Generate the corresponding Kappa for the PySB model associated with
        the exporter. A wrapper around
        :py:class:`pysb.generator.kappa.KappaGenerator`.

        Parameters
        ----------
        dialect : (optional) string, either 'kasim' (default) or 'complx'
            The Kappa file syntax for the Kasim simulator is slightly
            different from that of the complx analyzer. This argument
            specifies which type of Kappa to produce ('kasim' is the default).

        Returns
        -------
        string
            The Kappa output.
        """
        kappa_str = ''
        if self.docstring:
            kappa_str += '# ' + self.docstring.replace('\n', '\n# ') + '\n'
        gen = KappaGenerator(self.model, dialect=dialect)
        kappa_str += gen.get_content()
        return kappa_str




"""An implementation of the model from:

Model-based dissection of CD95 signaling dynamics reveals both a pro- and
antiapoptotic role of c-FLIPL.  Fricker N, Beaudouin J, Richter P, Eils R,
Krammer PH, Lavrik IN. J Cell Biol. 2010 Aug 9;190(3):377-89.
doi:10.1083/jcb.201002060

http://jcb.rupress.org/content/190/3/377.long

Implemented by: Jeremie Roux, Will Chen, Jeremy Muhlich
"""

from __future__ import print_function
from pysb import *

Model()

# Non-zero initial conditions (in molecules per cell):
Parameter('L_0'        , 1500e3); # baseline level of ligand for most experiments (corresponding to 50 ng/ml SuperKiller TRAIL)
Parameter('pR_0'       , 170.999e3);  # TRAIL receptor (for experiments not involving siRNA)
Parameter('FADD_0'     , 133.165e3);
Parameter('flipL_0'    , 0.49995e3);  # FlipL 1X = 0.49995e3
Parameter('flipS_0'    , 0.422e3);  # Flip
Parameter('pC8_0'      , 200.168e3);  # procaspase-8 (pro-C8)
Parameter('Bid_0'       , 100e3);  # Bid

Monomer('L', ['b'])
Monomer('pR', ['b', 'rf'])
Monomer('FADD', ['rf', 'fe'])
Monomer('flipL', ['b', 'fe', 'ee', 'D384'],
        {'D384': ['U','C']}
        )
Monomer('flipS', ['b', 'fe', 'ee'])
Monomer('pC8', ['fe', 'ee', 'D384', 'D400'],
        {'D384': ['U','C'],
	 'D400': ['U','C']}
        )
Monomer('Bid') #called Apoptosis substrat in Lavrik's model
Monomer('tBid')

flip_monomers = (flipL, flipS);

# L + R <--> L:R
Parameter('kf1', 70.98e-03) #70.98e-03
Parameter('kr1', 0)
Rule('R_L_Binding', L (b=None) + pR (b=None, rf=None) >> L (b=1) % pR (b=1, rf=None), kf1)

# FADD binds
Parameter('kf29', 84.4211e-03) #84.4211e-03
Parameter('kr29', 0)
Rule('RL_FADD_Binding', pR (b=ANY, rf=None) + FADD (rf=None, fe=None) >> pR (b=ANY, rf=2) % FADD (rf=2, fe=None), kf29,kr29)

#C8 binds to L:R:FADD
Parameter('kf30', 3.19838e-03) #3.19838e-03
Parameter('kr30', 0.1) #0.1
Rule('RLFADD_C8_Binding', FADD (rf=ANY, fe=None) + pC8 (fe=None, ee=None, D384='U') | FADD (rf=ANY, fe=1) % pC8 (fe=1, ee=None, D384='U'), kf30, kr30)

#FLIP(variants) bind to L:R:FADD
Parameter('kf31', 69.3329e-03)
Parameter('kr31', 0.0)
Parameter('kf32', 69.4022e-03)
Parameter('kr32', 0.08)
# FIXME: this pattern requires a dummy kr31 which is ultimately ignored
for flip_m, kf, kr, reversible in (zip(flip_monomers, (kf31,kf32), (kr31,kr32), (False,True))):
    rule = Rule('RLFADD_%s_Binding' % flip_m.name, FADD (rf=ANY, fe=None) + flip_m (fe=None, ee=None) | FADD (rf=ANY, fe=1) % flip_m (fe=1, ee=None), kf, kr)
    if reversible is False:
        rule.is_reversible = False
        rule.rule_expression.is_reversible = False
        rule.rate_reverse = None

pC8_HomoD   = pC8 (fe=ANY, ee=1, D384='U') % pC8   (fe=ANY, ee=1, D384='U')
pC8_HeteroD = pC8 (fe=ANY, ee=1, D384='U') % flipL (fe=ANY, ee=1, D384='U')
p43_HomoD   = pC8 (fe=ANY, ee=1, D384='C') % pC8   (fe=ANY, ee=1, D384='C')
p43_HeteroD = pC8 (fe=ANY, ee=1, D384='C') % flipL (fe=ANY, ee=1, D384='C')

#L:R:FADD:C8 dimerizes
Parameter('kf33', 2.37162)
Parameter('kr33', 0.1)
Parameter('kc33', 1e-05)
Rule('RLFADD_C8_C8_Binding', pC8 (fe=ANY, ee=None, D384='U') + pC8 (fe=ANY, ee=None, D384='U') | pC8_HomoD, kf33, kr33)

#L:R:FADD:C8 L:R:FADD:FLIP(variants) dimerizes
Parameter('kf34', 4.83692)
Parameter('kr34', 0)
Parameter('kf35', 2.88545)
Parameter('kr35', 1)
# FIXME: this pattern requires a dummy kr31 which is ultimately ignored
for flip_m, kf, kr, reversible in (zip(flip_monomers, (kf34,kf35), (kr34,kr35), (False,True))):
    rule = Rule('RLFADD_C8_%s_Binding' % flip_m.name, pC8 (fe=ANY, ee=None, D384='U') + flip_m (fe=ANY, ee=None) | pC8 (fe=ANY, ee=1, D384='U') % flip_m (fe=ANY, ee=1), kf, kr)
    if reversible is False:
        rule.is_reversible = False
        rule.rule_expression.is_reversible = False
        rule.rate_reverse = None

Parameter('kc36', 0.223046e-3)
#Homodimer catalyses Homodimer ?: no p18 is released. Only this "cleaved" p43 homoD is the product that will transform into a p18 + L:R:FADD in later reaction.
Rule('HomoD_cat_HomoD', pC8_HomoD + pC8_HomoD >> pC8_HomoD + p43_HomoD, kc36)
#Homodimer catalyses Heterodimer ?????
Rule('HomoD_cat_HeteroD', pC8_HomoD + pC8_HeteroD >> pC8_HomoD + p43_HeteroD, kc36)

Parameter('kc37', 0.805817e-3)
#Heterodimer catalyses Heterodimer ?????
Rule('HeteroD_cat_HeteroD', pC8_HeteroD + pC8_HeteroD >> pC8_HeteroD + p43_HeteroD, kc37)
#Heterodimer catalyses Homodimer ?????
Rule('HeteroD_cat_HomoD', pC8_HeteroD + pC8_HomoD >> pC8_HeteroD + p43_HomoD, kc37)

Parameter('kc38', 1.4888e-3)
#Cleaved Homodimer catalyses Homodimer ?????
Rule('Cl_HomoD_cat_HomoD', p43_HomoD + pC8_HomoD >> p43_HomoD + p43_HomoD, kc38)
#Cleaved HomoD catalyses Heterodimer ?????
Rule('Cl_HomoD_cat_HeteroD', p43_HomoD + pC8_HeteroD >> p43_HomoD + p43_HeteroD, kc38)

Parameter('kc39', 13.098e-3)
#Cleaved HeteroD catalyses Homodimer ?????
Rule('Cl_heteroD_cat_HomoD', p43_HeteroD + pC8_HomoD >> p43_HeteroD + p43_HomoD, kc39)
#Cleaved HeteroD catalyses Heterodimer ?????
Rule('Cl_heteroD_cat_HeteroD', p43_HeteroD + pC8_HeteroD >> p43_HeteroD + p43_HeteroD, kc39)

#Cleaved HomoD catalyses Cleaved HomoD to p18 and release L:R:FADD
Parameter('kc40', 0.999273e-3)
Rule('Cl_HomoD_cat_Cl_HomoD', pC8 (fe=ANY, ee=1, D384='C', D400='U') % pC8 (fe=ANY, ee=1, D384='C', D400='U') +
     FADD (rf=ANY, fe=2) % pC8 (fe=2, ee=3, D384='C', D400='U') % FADD (rf=ANY, fe=4) % pC8 (fe=4, ee=3, D384='C', D400='U') >>
     pC8 (fe=ANY, ee=1, D384='C', D400='U') % pC8 (fe=ANY, ee=1, D384='C', D400='U') +
     FADD (rf=ANY, fe=None) + FADD (rf=ANY, fe=None) + pC8 (fe=None, ee=1, D384='C',D400='C') % pC8 (fe=None, ee=1, D384='C',D400='C'), 
     kc40)

#Cleaved HeteroD catalyses Cleaved HomoD to p18 and release L:R:FADD
Parameter('kc41', 0.982109e-3)
Rule('Cl_HeteroD_cat_Cl_HomoD', pC8 (fe=ANY, ee=1, D384='C', D400='U') % flipL (fe=ANY, ee=1, D384='C') +
     FADD (rf=ANY, fe=2) % pC8 (fe=2, ee=3, D384='C', D400='U') % FADD (rf=ANY, fe=4) % pC8 (fe=4, ee=3, D384='C', D400='U') >>
     pC8 (fe=ANY, ee=1, D384='C', D400='U') % flipL (fe=ANY, ee=1, D384='C') +
     FADD (rf=ANY, fe=None) + FADD (rf=ANY, fe=None) + pC8 (fe=None, ee=1, D384='C',D400='C') % pC8 (fe=None, ee=1, D384='C',D400='C'), 
     kc41)
 
#Cleaved HomoD cleaves Bid ?????
Parameter('kc42', 0.0697394e-3)
Rule('Cl_Homo_cat_Bid', pC8 (fe=ANY, ee=1, D384='C', D400='U') % pC8 (fe=ANY, ee=1, D384='C', D400='U') + Bid () >>
     pC8 (fe=ANY, ee=1, D384='C', D400='U') % pC8 (fe=ANY, ee=1, D384='C', D400='U') + tBid (), kc42)

#Cleaved HeteroD cleaves Bid ?????
Parameter('kc43', 0.0166747e-3)
Rule('Cl_Hetero_cat_Bid', pC8 (fe=ANY, ee=1, D384='C', D400='U') % flipL (fe=ANY, ee=1, D384='C') + Bid () >>
     pC8 (fe=ANY, ee=1, D384='C', D400='U') % flipL (fe=ANY, ee=1, D384='C') + tBid (), kc43)

#p18 cleaves Bid ?????
Parameter('kc44', 0.0000479214e-3)
Rule('p18_Bid_cat', pC8 (fe=None, ee=1, D384='C',D400='C') % pC8 (fe=None, ee=1, D384='C',D400='C') + Bid () >> 
	pC8 (fe=None, ee=1, D384='C',D400='C') % pC8 (fe=None, ee=1, D384='C',D400='C') + tBid (), kc44) 


# Fig 4B

Observable('p18', pC8(fe=None, ee=1, D384='C',D400='C') % pC8(fe=None, ee=1, D384='C',D400='C'))
Observable('tBid_total', tBid() )




# generate initial conditions from _0 parameter naming convention
for m in model.monomers:
    ic_param = model.parameters.get('%s_0' % m.name, None)
    if ic_param is not None:
        sites = {}
        for s in m.sites:
            if s in m.site_states:
                sites[s] = m.site_states[s][0]
            else:
                sites[s] = None
        Initial(m(sites), ic_param)


####

if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")


####
# some of the model definition from the supplemental materials, for reference:

# ********** MODEL STATES
# %% Protein amounts are given in thousand molecules per cell.
# CD95L(0) = 1,500%% amount ligand
# CD95R(0) = 170.999%% amount CD95
# FADD(0) = 133.165%% amount FADD
# C8(0) = 200.168%% amount Procaspase-8
# FL(0) = 0.49995%% amount FLIP-Long
# FS(0) = 0.422%% amount FLIP-Short
# CD95RL(0) = 0%% amount of CD95-CD95L complexes
# CD95FADD(0) = 0%% amount of CD95-FADD complexes
# FADDC8(0) = 0%% amount Procaspase-8 bound to FADD
# FADDFL(0) = 0%% amount c-FLIPL bound to FADD
# FADDFS(0) = 0%% amount c-FLIPS bound to FADD
# C8heterodimer(0) = 0%% amount Procaspase-8/c-FLIPL heterodimers
# C8homodimer(0) = 0%% amount Procaspase-8 homodimers
# C8FSdimer(0) =0%% amount Procaspase-8/c-FLIPS heterodimers
# p43heterodimer(0) = 0%% amount p43/p41-Procaspase-8/p43-FLIP heterodimers
# p43homodimer(0) = 0%% amount p43/p41-Procaspase-8 homodimers
# p18(0)=0%% amount p18 formed
# apoptosissubstrate(0)=100
# cleavedsubstrate(0) = 0%% amount cleaved apoptosis substrate

# ********** MODEL VARIABLES
# p18total = 2 x p18
# p43Casp8total = 2 x p43homodimer + p43heterodimer
# procaspase8total = C8 + FADDC8 + C8heterodimer + 2 x C8homodimer + C8FSdimer
# c8total = p43Casp8total + procaspase8total + 2 x p18
# cleavedC8 = c8total - procaspase8total
# celldeath = cleavedsubstrate / 0.10875%% Model readout: percentage of dead cells

# ********** MODEL REACTIONS
# RCD95LBindCD95R = 7.0980e-002 x CD95L x CD95R
# RFADDBindCD95RL = 0.0844211 x CD95RL x FADD
# RC8BindCD95FADD = 0.00319838 x CD95FADD x C8
# RFLBindCD95FADD = 0.0693329 x CD95FADD x FL
# RFSBindCD95FADD = 0.0694022 x CD95FADD x FS
# RFADDC8Dissociate = 0.1 x FADDC8
# RFADDFSDissociate = 0.08 x FADDFS
# RFADDC8BindFADDC8 = 1.18581 x FADDC8 x FADDC8
# RFADDFLBindFADDC8 = 4.83692 x FADDC8 x FADDFL
# RFADDFSBindFADDC8 = 2.88545 x FADDC8 x FADDFS
# RC8FSdimerDissociate = 1 x C8FSdimer
# RC8homodimerDissociate = 0.1 x C8homodimer
# RC8homodimerCleaveC8homodimer = 0.000223046 x C8homodimer x C8homodimer
# RC8homodimerCleaveC8heterodimer = 0.000223046 x C8homodimer x C8heterodimer
# RC8heterodimerCleaveC8heterodimer = 0.000805817 x C8heterodimer x C8heterodimer
# RC8heterodimerCleaveC8homodimer = 0.000805817 x C8heterodimer x C8homodimer
# Rp43homodimerCleaveC8homodimer = 0.0014888 x p43homodimer x C8homodimer
# Rp43homodimerCleaveC8heterodimer = 0.0014888 x p43homodimer x C8heterodimer
# Rp43heterodimerCleaveC8homodimer = 0.013098 x p43heterodimer x C8homodimer
# Rp43heterodimerCleaveC8heterodimer = 0.013098 x p43heterodimer x C8heterodimer
# Rp43homodimerCleavep43homodimer = 0.000999273 x p43homodimer x p43homodimer
# Rp43heterodimerCleavep43homodimer = 0.000982109 x p43heterodimer x p43homodimer
# Rp43heterodimerCleaveApoptosisSubstrate = 1.66747e-005 x p43heterodimer x apoptosissubstrate
# Rp43homodimerCleaveApoptosisSubstrate = 6.97394e-005 x p43homodimer x apoptosissubstrate
# Rp18CleaveApoptosisSubstrate = 4.79214e-08 x p18 x apoptosissubstrate



"""Translation of the BioNetGen example "Enzymatic Cycle" from the BNG wiki.

http://bionetgen.org/index.php/Enzymatic_Cycle

This is the Michaelis-Menten implementation shown on the above webpage. I chose
this one simply because it exercises more PySB features than the basic
implementation.
"""

from pysb import *

Model()


# kinetic parameters
Parameter('k1', 1.0)    # kf   - units: /time/molecule
Parameter('k2', 1.0)    # kr   - units: /time
Parameter('k3', 0.1)    # kcat - units: /time
Parameter('k4', 1.0)    # kf   - units: /time/molecule
Parameter('k5', 1.0)    # kr   - units: /time
Parameter('k6', 0.1)    # kcat - units: /time
# total enzyme and substrate (units: molecules)
Parameter('E1_tot', 1)
Parameter('E4_tot', 1)
Parameter('S2_tot', 50)
Parameter('S5_tot', 50)
# derived parameters (units: molecules)
Expression('Km2', (k2 + k3) / k1)
Expression('Km5', (k5 + k6) / k4)
# calculate initial free S2 and S5, assuming quasi-equillibrium enzyme binding
Expression('S2_free0',
           (S2_tot + E1_tot - Km2) / 2 +
           ((S2_tot + E1_tot - Km2) ** 2 + 4 * Km2) ** 0.5 / 2)
Expression('S5_free0',
           (S5_tot + E4_tot - Km5) / 2 +
           ((S5_tot + E4_tot - Km5) ** 2 + 4 * Km5) ** 0.5 / 2)


Monomer('S', ['c'], {'c': ['c2', 'c5']})
Monomer('E1')
Monomer('E4')


Initial(E1(), E1_tot)           # S1+S3 (total enzyme A)
Initial(S(c='c2'), S2_free0)    # S2 (free substrate in config 2)
Initial(E4(), E4_tot)           # S4+S6 (total enzyme B)
Initial(S(c='c5'), S5_free0)    # S5 (free substrate in config 5)


# free substrate (required for functional ratelaw)
Observable('S2_free', S(c='c2'))
Observable('S5_free', S(c='c5'))


# s2 -> s5, catalyzed by E1
Expression('s2_s5_rate', k3 / (Km2 + S2_free))
Rule('s2_s5_cat_E1', S(c='c2') + E1() >> S(c='c5') + E1(), s2_s5_rate)
# s5 -> s2, catalyzed by E4
Expression('s5_s2_rate', k6 / (Km5 + S5_free))
Rule('s5_s2_cat_E4', S(c='c5') + E4() >> S(c='c2') + E4(), s5_s2_rate)



from pysb.simulator.base import Simulator, SimulationResult, SimulatorException
from pysb.generator.kappa import KappaGenerator
from pysb.kappa import KasimInterfaceError, _parse_kasim_outfile
import numpy as np
import os
import tempfile
import shutil
import pysb.pathfinder as pf
import subprocess


class KappaSimulator(Simulator):
    """ Simulate a model using Kappa """
    _supports = {
        'multi_initials':     True,
        'multi_param_values': True
    }

    def __init__(self, model, tspan=None, cleanup=True, verbose=False):
        super(KappaSimulator, self).__init__(model, tspan=tspan,
                                             verbose=verbose)
        self.cleanup = cleanup
        self._outdir = None

    def run(self, tspan=None, initials=None, param_values=None, n_runs=1,
            output_dir=None, output_file_basename=None,
            cleanup=None, **additional_args):
        """
        Simulate a model using Kappa

        Parameters
        ----------
        tspan: vector-like
            time span of simulation
        initials: vector-like, optional
            initial conditions of model
        param_values : vector-like or dictionary, optional
            Values to use for every parameter in the model. Ordering is
            determined by the order of model.parameters.
            If not specified, parameter values will be taken directly from
            model.parameters.
        n_runs: int
            number of simulations to run
        output_dir : string, optional
            Location for temporary files generated by Kappa. If None (the
            default), uses a temporary directory provided by the system. A
            temporary directory with a random name is created within the
            supplied location.
        output_file_basename : string, optional
            This argument is used as a prefix for the temporary Kappa
            output directory, rather than the individual files.
        cleanup : bool, optional
            If True (default), delete the temporary files after the
            simulation is finished. If False, leave them in place (Useful for
            debugging). The default value, None, means to use the value
            specified in :py:func:`__init__`.
        additional_args: kwargs, optional
            Additional arguments to pass to Kappa

                * seed : int, optional
                    Random number seed for Kappa simulation

                * perturbation : string, optional
                    Optional perturbation language syntax to be appended to the
                    Kappa file. See KaSim manual for more details.

        Examples
        --------

        >>> import numpy as np
        >>> from pysb.examples import michment
        >>> from pysb.simulator import KappaSimulator
        >>> sim = KappaSimulator(michment.model, tspan=np.linspace(0, 1))
        >>> x = sim.run(n_runs=1)


        """
        super(KappaSimulator, self).run(tspan=tspan,
                                        initials=initials,
                                        param_values=param_values,
                                        _run_kwargs=locals()
                                        )

        if cleanup is None:
            cleanup = self.cleanup

        tspan_lin_spaced = np.allclose(
            self.tspan,
            np.linspace(self.tspan[0], self.tspan[-1], len(self.tspan))
        )
        if not tspan_lin_spaced or self.tspan[0] != 0.0:
            raise SimulatorException('Kappa requires tspan to be linearly '
                                     'spaced starting at t=0')
        points = len(self.tspan)
        time = self.tspan[-1]
        plot_period = time / (len(self.tspan) - 1)

        if output_file_basename is None:
            output_file_basename = 'tmpKappa_%s_' % self.model.name

        base_directory = tempfile.mkdtemp(prefix=output_file_basename,
                                          dir=output_dir)

        base_filename = os.path.join(base_directory, self.model.name)
        kappa_filename_pattern = base_filename + '_{}.ka'
        out_filename_pattern = base_filename + '_{}_run{}.out'

        base_args = ['-u', 'time', '-l', str(time), '-p', '%.5f' % plot_period]
        if 'seed' in additional_args:
            seed = additional_args.pop('seed')
            base_args.extend(['-seed', str(seed)])

        kasim_path = pf.get_path('kasim')
        n_param_sets = self.initials_length

        gen = KappaGenerator(self.model, _exclude_ic_param=True)
        file_data_base = gen.get_content()

        # Check if a perturbation has been set
        try:
            perturbation = additional_args.pop('perturbation')
        except KeyError:
            perturbation = None

        # Check no unknown arguments have been set
        if additional_args:
            raise ValueError('Unknown argument(s): {}'.format(
                ', '.join(additional_args.keys())
            ))

        # Kappa column names, for sanity check
        kappa_col_names = tuple(['time'] +
                                [o.name for o in self.model.observables])
        tout = []
        observable_traj = []
        try:
            for pset_idx in range(n_param_sets):
                file_data = file_data_base + ''
                for param, param_value in zip(self.model.parameters,
                                              self.param_values[pset_idx]):
                    file_data += "%var: '{}' {:e}\n".format(param.name,
                                                            param_value)
                file_data += '\n'
                for cp, values in self.initials_dict.items():
                    file_data += "%init: {} {}\n".format(
                        values[pset_idx], gen.format_complexpattern(cp)
                    )

                # If any perturbation language code has been passed in, add it
                # to the Kappa file:
                if perturbation:
                    file_data += '%s\n' % perturbation

                # Generate the Kappa model code from the PySB model and write
                # it to the Kappa file:
                kappa_filename = kappa_filename_pattern.format(pset_idx)
                with open(kappa_filename, 'w') as kappa_file:
                    self._logger.debug('Kappa file contents:\n\n' + file_data)
                    kappa_file.write(file_data)

                for sim_rpt in range(n_runs):
                    # Run Kappa
                    out_filename = out_filename_pattern.format(pset_idx,
                                                               sim_rpt)
                    args = [kasim_path] + base_args + [
                        '-i', kappa_filename, '-o', out_filename]

                    # Run KaSim
                    self._logger.debug('Running: {}'.format(' '.join(args)))
                    p = subprocess.Popen(args,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         cwd=base_directory)

                    for line in p.stdout:
                        self._logger.debug('@@' + line.decode('utf8')[:-1])
                    (p_out, p_err) = p.communicate()

                    if p.returncode:
                        raise KasimInterfaceError(
                            p_out.decode('utf8') + '\n' + p_err.decode('utf8'))

                    # The simulation data, as a numpy array
                    data = _parse_kasim_outfile(out_filename)
                    # Sanity check that observables are in correct order
                    assert data.dtype.names == kappa_col_names
                    data = data.view('<f8')
                    # Handle case with single row output
                    if data.ndim == 1:
                        data.shape = (1, data.shape[0])
                    # Parse into format
                    tout.append(data[:, 0])
                    observable_traj.append(data[:, 1:])
        finally:
            if cleanup:
                shutil.rmtree(base_directory)

        return SimulationResult(self, tout=tout,
                                observables_and_expressions=observable_traj,
                                simulations_per_param_set=n_runs)



from pysb.core import MonomerPattern, ComplexPattern, RuleExpression, \
    ReactionPattern, ANY, WILD, MultiState
from pysb.builder import Builder
from pysb.bng import BngFileInterface, parse_bngl_expr
import xml.etree.ElementTree
import re
import sympy
from sympy.parsing.sympy_parser import parse_expr
import warnings
import pysb.logging
import collections
import numbers
import os


def _ns(tag_string):
    """
    Shorthand to apply XML namespace
    """
    return tag_string.format('{http://www.sbml.org/sbml/level3}')


class BnglImportError(Exception):
    pass


class BnglBuilder(Builder):
    """
    Assemble a Model from a .bngl file.

    See :py:func:`model_from_bngl` for further details.
    """
    def __init__(self, filename, force=False, cleanup=True):
        super(BnglBuilder, self).__init__()
        filename = os.path.abspath(filename)
        with BngFileInterface(model=None, cleanup=cleanup) as con:
            con.action('readFile', file=filename, skip_actions=1)
            con.action('writeXML', evaluate_expressions=0)
            con.execute()
            self._x = xml.etree.ElementTree.parse('%s.xml' %
                                                  con.base_filename)\
                                           .getroot().find(_ns('{0}model'))

        self.model.name = os.path.splitext(os.path.basename(filename))[0]
        self._force = force
        self._model_env = {}
        self._renamed_states = collections.defaultdict(dict)
        self._log = pysb.logging.get_logger(__name__)
        self._parse_bng_xml()

    def _warn_or_except(self, msg):
        """
        Raises a warning or Exception, depending of the value of self._force
        """
        if self._force:
            warnings.warn(msg)
        else:
            raise BnglImportError(msg)

    def _eval_in_model_env(self, expression):
        """
        Evaluates an expression string in the model environment
        """
        components = dict((c.name, c) for c in self.model.all_components())
        self._model_env.update(components)

        # Quick security check on the expression
        if re.match(r'^[\w\s()/+\-._*]*$', expression):
            return parse_bngl_expr(expression, local_dict=self._model_env,
                                   evaluate=False)
        else:
            self._warn_or_except('Security check on expression "%s" failed' %
                                 expression)
            return None

    def _parse_species(self, species_xml):
        # Species may be empty for synthesis/degradation reaction patterns
        if species_xml is None:
            return []
        # Give the bonds unique numeric IDs
        bond_ids = {}
        next_bond_id = 1
        for bond in species_xml.iterfind(_ns('{0}ListOfBonds/{0}Bond')):
            for bond_attr in bond.attrib.keys():
                if bond_attr.startswith('site'):
                    bond_ids.setdefault(bond.attrib[bond_attr], [])\
                                        .append(next_bond_id)
            next_bond_id += 1

        # Create a list of monomer patterns
        mon_pats = []
        for mon in species_xml.iterfind(_ns('{0}ListOfMolecules/{0}Molecule')):
            mon_name = mon.get('name')
            mon_obj = self.model.monomers[mon_name]
            mon_states = collections.defaultdict(list)
            for comp in mon.iterfind(_ns('{0}ListOfComponents/{0}Component')):
                # BioNetGen component (state) labels are not supported yet
                if 'label' in comp.attrib.keys():
                    self._warn_or_except('BioNetGen component/state labels '
                                         'are not yet supported in PySB')

                state_nm = comp.get('name')
                bonds = comp.get('numberOfBonds')
                if bonds == "0":
                    last_bond = None
                elif bonds == "?":
                    last_bond = WILD
                elif bonds == "+":
                    last_bond = ANY
                else:
                    bond_list = bond_ids[comp.get('id')]
                    assert int(bonds) == len(bond_list)
                    if len(bond_list) == 1:
                        last_bond = bond_list[0]
                    else:
                        last_bond = bond_list
                state = comp.get('state')
                if state and state != '?':
                    if state in ('PLUS', 'MINUS'):
                        self._warn_or_except(
                            'PLUS/MINUS state values are non-standard BNGL '
                            'used to increment or decrement a numeric state '
                            'value. They are not supported in PySB.'
                        )
                    # If we changed the state string, use the updated version
                    state = self._renamed_states.get(mon_name, {}).get(
                        state, state)
                    if last_bond:
                        # Site has a bond and a state
                        mon_states[state_nm].append((state, last_bond))
                    else:
                        # Site only has a state, no bond
                        mon_states[state_nm].append(state)
                else:
                    mon_states[state_nm].append(last_bond)

            mon_states = {k: MultiState(*v) if len(v) > 1 else v[0]
                          for k, v in mon_states.items()}

            mon_cpt = self.model.compartments.get(mon.get('compartment'))
            mon_pats.append(MonomerPattern(mon_obj, mon_states, mon_cpt))
            if 'label' in mon.attrib.keys():
                try:
                    tag = self.model.tags[mon.get('label')]
                except KeyError:
                    tag = self.tag(mon.get('label'))
                mon_pats[-1]._tag = tag
        return mon_pats

    def _parse_monomers(self):
        for m in self._x.iterfind(_ns('{0}ListOfMoleculeTypes/'
                                      '{0}MoleculeType')):
            mon_name = m.get('id')
            sites = []
            states = collections.defaultdict(dict)
            for ctype in m.iterfind(_ns('{0}ListOfComponentTypes/'
                                        '{0}ComponentType')):
                c_name = ctype.get('id')
                sites.append(c_name)
                states_list = ctype.find(_ns('{}ListOfAllowedStates'))
                if states_list is not None:
                    for s in states_list.iterfind(_ns('{}AllowedState')):
                        state = s.get('id')
                        if re.match('[0-9]*$', state):
                            if state not in states[c_name]:
                                new_state = '_' + state
                                while new_state in states[c_name].values():
                                    new_state = '_' + new_state
                                states[c_name][state] = new_state
                                self._renamed_states[mon_name][state] = \
                                    new_state
                        else:
                            states[c_name][state] = state
                    if self._renamed_states[mon_name]:
                        self._log.info('Monomer "{}" states were renamed as '
                                       'follows: {}'.format(
                            mon_name,
                            self._renamed_states[mon_name])
                        )
            try:
                self.monomer(mon_name, sites,
                             {c_name: list(statedict.values())
                              for c_name, statedict in states.items()})
            except Exception as e:
                if str(e).startswith('Duplicate sites specified'):
                    self._warn_or_except('Molecule %s has multiple '
                                         'sites with the same name. '
                                         'This is not supported in PySB.' %
                                         mon_name)
                else:
                    raise BnglImportError(str(e))

    def _parse_parameters(self):
        for p in self._x.iterfind(_ns('{0}ListOfParameters/{0}Parameter')):
            p_name = p.get('id')
            if p.get('type') == 'Constant':
                p_value = p.get('value').replace('10^', '1e')
                try:
                    self.parameter(name=p_name, value=p_value)
                except ValueError:
                    # Despite the "Constant" label, some constant expressions
                    # appear here e.g. ln(2)/120 in BNG's Repressilator model
                    self.expression(name=p_name,
                                    expr=self._eval_in_model_env(
                                        p.get('value')))
            elif p.get('type') == 'ConstantExpression':
                self.expression(name=p_name,
                                expr=self._eval_in_model_env(p.get('value')))
            else:
                self._warn_or_except('Parameter %s has unknown type: %s' %
                                     (p_name, p.get('type')))

    def _parse_observables(self):
        for o in self._x.iterfind(_ns('{0}ListOfObservables/{0}Observable')):
            o_name = o.get('name')
            match_mode = o.get('type').lower()
            # Some BNG observables have same name as a monomer, but in PySB
            # these must be unique
            if o_name in self.model.monomers.keys():
                o_name = 'Obs_{}'.format(o_name)
            cplx_pats = []
            for mp in o.iterfind(_ns('{0}ListOfPatterns/{0}Pattern')):
                cpt = self.model.compartments.get(mp.get('compartment'))
                match_once = mp.get('matchOnce')
                match_once = True if match_once == "1" and \
                    match_mode != 'species' else False
                cplx_pats.append(ComplexPattern(self._parse_species(mp),
                                                compartment=cpt,
                                                match_once=match_once))
            self.observable(o_name,
                            ReactionPattern(cplx_pats),
                            match=match_mode)

    def _parse_initials(self):
        for i in self._x.iterfind(_ns('{0}ListOfSpecies/{0}Species')):
            value_param = i.get('concentration')
            try:
                value = float(value_param)
                # Need to create a new parameter for the initial conc. literal
                name = re.sub('[^\w]+', '_', i.get('name').replace(
                    ')', '').replace('(', '')) + '_0'
                try:
                    value_param = self.parameter(name, value)
                except ValueError as ve:
                    raise BnglImportError(str(ve))
            except ValueError:
                # Retrieve existing parameter or (constant) expression
                try:
                    value_param = self.model.parameters[i.get('concentration')]
                except KeyError:
                    value_param = self.model.expressions[i.get(
                        'concentration')]
            mon_pats = self._parse_species(i)
            species_cpt = self.model.compartments.get(i.get('compartment'))
            cp = ComplexPattern(mon_pats, species_cpt)
            fixed = i.get('Fixed') == "1"
            self.initial(cp, value_param, fixed)

    def _parse_compartments(self):
        for c in self._x.iterfind(_ns('{0}ListOfCompartments/{0}compartment')):
            cpt_size = None
            if c.get('size'):
                cpt_size = self.parameter('%s_size' % c.get('id'),
                                          c.get('size'))
            cpt_parent = None
            if c.get('outside'):
                cpt_parent = self.model.compartments[c.get('outside')]
            self.compartment(name=c.get('id'),
                             parent=cpt_parent,
                             dimension=int(c.get('spatialDimensions')),
                             size=cpt_size)

    def _parse_rate_law(self, rl):
        if rl.get('type') == 'Ele':
            rate_law_name = rl.find(_ns('{0}ListOfRateConstants/'
                                        '{0}RateConstant')).get('value')
            try:
                return self.model.parameters[rate_law_name]
            except KeyError:
                return self.model.expressions[rate_law_name]
        elif rl.get('type') == 'Function':
            try:
                return self.model.expressions[rl.get('name')]
            except KeyError:
                return self.model.parameters[rl.get('name')]
        else:
            self._warn_or_except('Rate law %s has unknown type %s' %
                                 (rl.get('id'), rl.get('type')))
            return None

    def _parse_rules(self):
        # Store reversible rates for post-processing (we don't know if we'll
        # encounter fwd or rev rule first)
        rev_rates = {}
        for r in self._x.iterfind(_ns('{0}ListOfReactionRules/'
                                      '{0}ReactionRule')):
            r_name = r.get('name')

            r_rate_xml = r.find(_ns('{}RateLaw'))
            if r_rate_xml is None:
                raise BnglImportError('Rate law missing for rule %s' % r_name)
            r_rate = self._parse_rate_law(r_rate_xml)

            # Store reverse rates for post-processing
            if r_name.startswith('_reverse_'):
                r_name = r_name[9:]
                rev_rates[r_name] = r_rate
                continue

            # Compile reactant and product patterns
            reactant_pats = []
            for rp in r.iterfind(_ns('{0}ListOfReactantPatterns/'
                                     '{0}ReactantPattern')):
                cpt = self.model.compartments.get(rp.get('compartment'))
                reactant_pats.append(ComplexPattern(self._parse_species(rp),
                                                    cpt))
                if 'label' in rp.attrib:
                    reactant_pats[-1]._tag = self.model.tags[rp.get('label')]
            product_pats = []
            for pp in r.iterfind(_ns('{0}ListOfProductPatterns/'
                                     '{0}ProductPattern')):
                cpt = self.model.compartments.get(pp.get('compartment'))
                product_pats.append(ComplexPattern(self._parse_species(pp),
                                                   cpt))

                if 'label' in pp.attrib:
                    product_pats[-1]._tag = self.model.tags[pp.get('label')]
            rule_exp = RuleExpression(ReactionPattern(reactant_pats),
                                      ReactionPattern(product_pats),
                                      is_reversible=False)

            # Process any DeleteMolecules declaration
            delete_molecules = False
            for del_operations in r.iterfind(_ns('{0}ListOfOperations/'
                                                 '{0}Delete')):
                if del_operations.get('DeleteMolecules') == "1":
                    delete_molecules = True
                    break

            # Process any MoveConnected declaration
            move_connected = False
            for change_cpt_ops in r.iterfind(_ns('{0}ListOfOperations/'
                                                 '{0}ChangeCompartment')):
                if change_cpt_ops.get('moveConnected') == "1":
                    move_connected = True
                    break

            # Give warning/error if ListOfExcludeReactants or
            # ListOfExcludeProducts is present
            if r.find(_ns('{}ListOfExcludeReactants')) is not None or \
               r.find(_ns('{}ListOfExcludeProducts')) is not None:
                self._warn_or_except('ListOfExcludeReactants and/or '
                                     'ListOfExcludeProducts declarations '
                                     'are deprecated in BNG, and not supported '
                                     'in PySB.')

            # Give warning/error if ListOfIncludeReactants or
            # ListOfIncludeProducts is present
            if r.find(_ns('{}ListOfIncludeReactants')) is not None or \
               r.find(_ns('{}ListOfIncludeProducts')) is not None:
                self._warn_or_except('ListOfIncludeReactants and/or '
                                     'ListOfIncludeProducts declarations '
                                     'are deprecated in BNG, and not supported '
                                     'in PySB.')

            self.rule(r_name, rule_exp, r_rate,
                      delete_molecules=delete_molecules,
                      move_connected=move_connected)

        # Set the reverse rates
        for r_name, rev_rate in rev_rates.items():
            rule = self.model.rules[r_name]
            rule.rule_expression.is_reversible = True
            rule.is_reversible = True
            rule.rate_reverse = rev_rate

    def _parse_expressions(self):
        expr_namespace = (
            self.model.parameters | self.model.expressions
            | self.model.observables
        )
        expr_symbols = {e.name: sympy.Symbol(e.name) for e in expr_namespace}

        for e in self._x.iterfind(_ns('{0}ListOfFunctions/{0}Function')):
            for arg in e.iterfind(_ns('{0}ListOfArguments/{0}Argument')):
                tag_name = arg.get('id')
                try:
                    self.model.tags[tag_name]
                except KeyError:
                    tag = self.tag(tag_name)
                    expr_symbols[tag_name] = tag
            expr_name = e.get('id')
            expr_text = e.find(_ns('{0}Expression')).text
            expr_val = 0
            try:
                expr_val = parse_bngl_expr(expr_text, local_dict=expr_symbols)
            except Exception as ex:
                self._warn_or_except('Could not parse expression %s: '
                                     '%s\n\nError: %s' % (expr_name,
                                                          expr_text,
                                                          str(ex)))
            if isinstance(expr_val, numbers.Number):
                self.parameter(expr_name, expr_val)
            else:
                self.expression(expr_name, expr_val)

    def _parse_bng_xml(self):
        self._parse_monomers()
        self._parse_parameters()
        self._parse_compartments()
        self._parse_initials()
        self._parse_observables()
        self._parse_expressions()
        self._parse_rules()


def model_from_bngl(filename, force=False, cleanup=True):
    """
    Convert a BioNetGen .bngl model file into a PySB Model.

    Notes
    -----

    The following features are not supported in PySB and will cause an error
    if present in a .bngl file:

    * Fixed species (with a ``$`` prefix, like ``$Null``)
    * BNG excluded or included reaction patterns (deprecated in BNG)
    * BNG local functions
    * Molecules with identically named sites, such as ``M(l,l)``
    * BNG's custom rate law functions, such as ``MM`` and ``Sat``
      (deprecated in BNG)

    Parameters
    ----------
    filename : string
        A BioNetGen .bngl file
    force : bool, optional
        The default, False, will raise an Exception if there are any errors
        importing the model to PySB, e.g. due to unsupported features.
        Setting to True will attempt to ignore any import errors, which may
        lead to a model that only poorly represents the original. Use at own
        risk!
    cleanup : bool
        Delete temporary directory on completion if True. Set to False for
        debugging purposes.
    """
    bb = BnglBuilder(filename, force=force, cleanup=cleanup)
    return bb.model



from pysb.core import SelfExporter, Component, Model

class Annotation(object):

    """
    A lightweight annotation mechanism for model elements.

    Based loosely on MIRIAM (http://co.mbine.org/standards/miriam) which is in
    turn based on RDF. An Annotation is equivalent to an RDF triple.

    This is still an experimental feature!

    Parameters
    ----------
    subject
        Element to annotate, typically a Component.
    object_
        Annotation, typically a string containing an identifiers.org URL.
    predicate : string, optional
        Relationship of `object_` to `subject`, typically a string containing a
        biomodels.net qualifier. If not specified, defaults to 'is'.

    """

    def __init__(self, subject, object_, predicate="is"):
        self.subject = subject
        self.object = object_
        self.predicate = predicate
        # if SelfExporter is in use, add the annotation to the model
        if SelfExporter.do_export:
            SelfExporter.default_model.add_annotation(self)

    def __repr__(self):
        if isinstance(self.subject, Component):
            subject = self.subject.name
        elif isinstance(self.subject, Model):
            subject = 'model'
        else:
            subject = self.subject
        return "%s(%s, %s, %s)" % (self.__class__.__name__, subject,
                                   repr(self.object), repr(self.predicate))



from .base import SimulatorException, SimulationResult
from .scipyode import ScipyOdeSimulator
from .cupsoda import CupSodaSimulator
from .stochkit import StochKitSimulator
from .bng import BngSimulator, PopulationMap
from .kappa import KappaSimulator

__all__ = ['BngSimulator', 'CupSodaSimulator', 'ScipyOdeSimulator',
           'StochKitSimulator', 'KappaSimulator',
           'SimulationResult', 'PopulationMap']



from __future__ import absolute_import
from __future__ import print_function as _
import pysb.core
from pysb.generator.bng import BngGenerator, format_complexpattern
import os
import subprocess
import re
import itertools
import sympy
import sympy.parsing.sympy_parser as sympy_parser
import numpy
import tempfile
import abc
from warnings import warn
import shutil
import collections
import pysb.pathfinder as pf
import tokenize
from pysb.logging import get_logger, EXTENDED_DEBUG

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from future_builtins import zip
except ImportError:
    pass

# Alias basestring under Python 3 for forwards compatibility
try:
    basestring
except NameError:
    basestring = str


def set_bng_path(dir):
    """ Deprecated. Use pysb.pathfinder.set_path() instead. """
    warn("Function %s() is deprecated; use pysb.pathfinder.set_path() "
         "instead" % set_bng_path.__name__, category=DeprecationWarning,
         stacklevel=2)
    pf.set_path('bng', dir)


class BngInterfaceError(RuntimeError):
    """BNG reported an error"""
    pass


class BngBaseInterface(object):
    """ Abstract base class for interfacing with BNG """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, model=None, verbose=False, cleanup=False,
                 output_prefix=None, output_dir=None,
                 model_additional_species=None,
                 model_population_maps=None):
        self._logger = get_logger(__name__,
                                  model=model,
                                  log_level=verbose)
        self._base_file_stem = 'pysb'
        self.cleanup = cleanup
        self.output_prefix = 'tmpBNG' if output_prefix is None else \
            output_prefix
        if model:
            self.generator = BngGenerator(
                model, additional_initials=model_additional_species,
                population_maps=model_population_maps
            )
            self.model = self.generator.model
            self._check_model()
        else:
            self.generator = None
            self.model = None

        self.base_directory = tempfile.mkdtemp(prefix=self.output_prefix,
                                               dir=output_dir)
        self._logger.debug('{} instantiated in directory {}'.format(
            self.__class__, self.base_directory)
        )

    def __enter__(self):
        return self

    @abc.abstractmethod
    def __exit__(self):
        return

    def _delete_tmpdir(self):
        shutil.rmtree(self.base_directory)

    def _check_model(self):
        """
        Checks a model has at least one initial condition and rule, raising
        an exception if not
        """
        if not self.model.rules:
            raise NoRulesError()
        if (
            not self.model.initials
            and not any(r.is_synth() for r in self.model.rules)
        ):
            raise NoInitialConditionsError()

    @classmethod
    def _bng_param(cls, param):
        """
        Ensures a BNG console parameter is in the correct format

        Strings are double quoted and booleans are mapped to [0,1]. Other
        types are currently used verbatim.

        Parameters
        ----------
        param :
            An argument to a BNG action call
        """
        if isinstance(param, basestring):
            return '"%s"' % param
        elif isinstance(param, bool):
            return 1 if param else 0
        elif isinstance(param, (collections.Sequence, numpy.ndarray)):
            return list(param)
        return param

    @abc.abstractmethod
    def action(self, action, **kwargs):
        """
        Generates code to execute a BNG action command

        Parameters
        ----------
        action: string
            The name of the BNG action function
        kwargs: kwargs, optional
            Arguments and values to supply to BNG
        """
        return

    @classmethod
    def _format_action_args(cls, **kwargs):
        """
        Formats a set of arguments for BNG

        Parameters
        ----------
        kwargs: kwargs, optional
            Arguments and values to supply to BNG
        """
        if kwargs:
            action_args = ','.join('%s=>%s' % (k, BngConsole._bng_param(v))
                                   for k, v in kwargs.items())
        else:
            action_args = ''
        return action_args

    @property
    def base_filename(self):
        """
        Returns the base filename (without extension) for BNG output files
        """
        return os.path.join(self.base_directory, self._base_file_stem)

    @property
    def bng_filename(self):
        """
        Returns the BNG command list (.bngl) filename (does not check
        whether the file exists)
        """
        return self.base_filename + '.bngl'

    @property
    def net_filename(self):
        """
        Returns the BNG network filename (does not check whether the file
        exists)
        """
        return self.base_filename + '.net'

    def read_netfile(self):
        """
        Reads a BNG network file as a string. Note that you must execute
        network generation separately before attempting this, or the file will
        not be found.
        :return: Contents of the BNG network file as a string
        """
        self._logger.debug('Reading BNG netfile: %s' % self.net_filename)
        with open(self.net_filename, 'r') as net_file:
            output = net_file.read()

        return output

    def read_simulation_results(self):
        """
        Read the results of a BNG simulation as a numpy array

        Returns
        -------
        numpy.ndarray
            Simulation results in a 2D matrix (time on Y axis,
            species/observables/expressions on X axis depending on
            simulation type)
        """
        return self.read_simulation_results_multi([self.base_filename])[0]

    @staticmethod
    def read_simulation_results_multi(base_filenames):
        """
        Read the results of multiple BNG simulations

        Parameters
        ----------
        base_filenames: list of str
            A list of filename stems to read simulation results in from,
            including the full path but not including any file extension.

        Returns
        -------
        list of numpy.ndarray
            List of simulation results, each in a 2D matrix (time on Y axis,
            species/observables/expressions on X axis depending on
            simulation type)
        """
        list_of_yfulls = []
        for base_filename in base_filenames:
            names = ['time']

            # Read concentrations data
            try:
                cdat_arr = numpy.loadtxt(base_filename + '.cdat', skiprows=1, ndmin=2)
                # -1 for time column
                names += ['__s%d' % i for i in range(cdat_arr.shape[1] - 1)]
            except IOError:
                cdat_arr = None

            # Read groups data
            try:
                with open(base_filename + '.gdat', 'r') as f:
                    # Exclude \# and time column
                    names += f.readline().split()[2:]
                    # Exclude first column (time)
                    gdat_arr = numpy.loadtxt(f, ndmin=2)
                    if cdat_arr is None:
                        cdat_arr = numpy.ndarray((len(gdat_arr), 0))
                    else:
                        gdat_arr = gdat_arr[:, 1:]
            except IOError:
                if cdat_arr is None:
                    raise BngInterfaceError('Need at least one of .cdat file or '
                                            '.gdat file to read simulation '
                                            'results')
                gdat_arr = numpy.ndarray((len(cdat_arr), 0))

            yfull_dtype = list(zip(names, itertools.repeat(float)))
            yfull = numpy.ndarray(len(cdat_arr), yfull_dtype)

            yfull_view = yfull.view(float).reshape(len(yfull), -1)
            yfull_view[:, :cdat_arr.shape[1]] = cdat_arr
            yfull_view[:, cdat_arr.shape[1]:] = gdat_arr

            list_of_yfulls.append(yfull)

        return list_of_yfulls


class BngConsole(BngBaseInterface):
    """ Interact with BioNetGen through BNG Console """
    def __init__(self, model=None, verbose=False, cleanup=True,
                 output_dir=None, output_prefix=None, timeout=30,
                 suppress_warnings=False, model_additional_species=None):
        super(BngConsole, self).__init__(
            model, verbose, cleanup, output_prefix, output_dir,
            model_additional_species=model_additional_species
        )

        try:
            import pexpect
        except ImportError:
            raise ImportError("Library 'pexpect' is required to use "
                              "BNGConsole, please install it to continue.\n"
                              "It is not currently available on Windows.")

        if suppress_warnings:
            warn("suppress_warnings is deprecated and has no effect. Adjust "
                 "the log level with the verbose argument instead.",
                 category=DeprecationWarning,
                 stacklevel=2)

        # Generate BNGL file
        if self.model:
            with open(self.bng_filename, mode='w') as bng_file:
                bng_file.write(self.generator.get_content())

        # Start BNG Console and load BNGL
        bng_path = pf.get_path('bng')
        bng_exec_path = '%s --console' % bng_path
        if not bng_path.endswith('.bat'):
            bng_exec_path = 'perl %s' % bng_exec_path
        self.console = pexpect.spawn(bng_exec_path,
                                     cwd=self.base_directory,
                                     timeout=timeout)
        self._console_wait()
        if self.model:
            self.console.sendline('load %s' % self.bng_filename)
            self._console_wait()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        In console mode, commands have already been executed, so we simply
        close down the console and erase the temporary directory if applicable.
        """
        self.console.sendline('done')
        self.console.close()
        if self.cleanup:
            self._delete_tmpdir()

    def _console_wait(self):
        """
        Wait for BNG console to process the command, and return the output
        :return: BNG console output from the previous command
        """
        self.console.expect('BNG>')
        # Python 3 requires explicit conversion of 'bytes' to 'str'
        console_msg = self.console.before.decode('utf-8')
        if "ERROR:" in console_msg:
            raise BngInterfaceError(console_msg)
        elif "WARNING:" in console_msg:
            self._logger.warning(console_msg)
        else:
            self._logger.debug(console_msg)
        return console_msg

    def generate_network(self, overwrite=False):
        """
        Generates a network in BNG and returns the network file contents as
        a string

        Parameters
        ----------
        overwrite: bool, optional
            Overwrite existing network file, if any
        """
        self.action('generate_network', overwrite=overwrite)
        return self.read_netfile()

    def action(self, action, **kwargs):
        """
        Generates a BNG action command and executes it through the console,
        returning any console output

        Parameters
        ----------
        action : string
            The name of the BNG action function
        kwargs : kwargs, optional
            Arguments and values to supply to BNG
        """
        # Process BNG arguments into a string
        action_args = self._format_action_args(**kwargs)

        # Execute the command via the console
        if action_args == '':
            cmd = 'action %s()' % action
        else:
            cmd = 'action %s({%s})' % (action, action_args)
        self._logger.debug(cmd)
        self.console.sendline(cmd)

        # Wait for the command to execute and return the result
        return self._console_wait()

    def load_bngl(self, bngl_file):
        """
        Load a BNGL file in the BNG console

        Parameters
        ----------
        bngl_file : string
            The filename of a .bngl file
        """
        cmd = 'load %s' % bngl_file
        self._logger.debug(cmd)
        self.console.sendline(cmd)
        self._console_wait()
        self._base_file_stem = os.path.splitext(os.path.basename(bngl_file))[0]


class BngFileInterface(BngBaseInterface):
    def __init__(self, model=None, verbose=False, output_dir=None,
                 output_prefix=None, cleanup=True,
                 model_additional_species=None,
                 model_population_maps=None):
        super(BngFileInterface, self).__init__(
            model, verbose, cleanup, output_prefix, output_dir,
            model_additional_species=model_additional_species,
            model_population_maps=model_population_maps
        )
        self._init_command_queue()

    def _init_command_queue(self):
        """
        Initializes the BNG command queue
        """
        self.command_queue = StringIO()
        self.command_queue.write('begin actions\n')

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        In file interface mode, we close the command queue buffer (whether
        or not it's been executed) and erase the temporary directory if
        applicable.
        """
        self.command_queue.close()
        if self.cleanup:
            self._delete_tmpdir()

    def execute(self, reload_netfile=False, skip_file_actions=True):
        """
        Executes all BNG commands in the command queue.

        Parameters
        ----------
        reload_netfile: bool or str
            If true, attempts to reload an existing .net file from a
            previous execute() iteration. If a string, the filename
            specified in the string is supplied to BNG's readFile (which can be
            any file type BNG supports, such as .net or .bngl).
            This is useful for running multiple actions in a row,
            where results need to be read into PySB before a new series of
            actions is executed.
        skip_file_actions: bool
            Only used if the previous argument is not False. Set this
            argument to True to ignore any actions block in the loaded file.
        """
        self.command_queue.write('end actions\n')
        bng_commands = self.command_queue.getvalue()

        # Generate BNGL file
        with open(self.bng_filename, 'w') as bng_file:
            output = ''
            if self.model and not reload_netfile:
                output += self.generator.get_content()
            if reload_netfile:
                filename = reload_netfile if \
                    isinstance(reload_netfile, basestring) \
                    else self.net_filename
                output += '\n  readFile({file=>"%s",skip_actions=>%d})\n' \
                    % (filename, int(skip_file_actions))
            output += bng_commands
            bng_file.write(output)
            lines = output.split('\n')
            line_number_format = 'L{{:0{}d}}  {{}}'.format(int(numpy.ceil(numpy.log10(len(lines)))))
            output = '\n'.join(line_number_format.format(ln + 1, line) for ln, line in enumerate(lines))
            self._logger.debug('BNG command file contents:\n' + output)

        # Reset the command queue, in case execute() is called again
        self.command_queue.close()
        self._init_command_queue()

        bng_exec_args = [pf.get_path('bng'), self.bng_filename]
        if not bng_exec_args[0].endswith('.bat'):
            bng_exec_args.insert(0, 'perl')

        p = subprocess.Popen(bng_exec_args,
                             cwd=self.base_directory,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

        # output lines as DEBUG, unless a warning or error is encountered
        capture_error = False
        captured_error_lines = []
        for line in iter(p.stdout.readline, b''):
            line = line[:-1].decode('utf-8')
            if line.startswith('ERROR:'):
                capture_error = True
            if capture_error:
                captured_error_lines.append(line)

            self._logger.debug(line)

        # p_out is already consumed, so only get p_err
        (_, p_err) = p.communicate()
        p_err = p_err.decode('utf-8')
        if p.returncode or captured_error_lines:
            raise BngInterfaceError('\n'.join(captured_error_lines) + "\n" +
                                    p_err.rstrip())

    def action(self, action, **kwargs):
        """
        Generates a BNG action command and adds it to the command queue

        Parameters
        ----------
        action : string
            The name of the BNG action function
        kwargs : kwargs, optional
            Arguments and values to supply to BNG
        """
        # Process BNG arguments into a string
        action_args = self._format_action_args(**kwargs)

        # Add the command to the queue
        if action_args == '':
            self.command_queue.write('\t%s()\n' % action)
        else:
            self.command_queue.write('\t%s({%s})\n' % (action, action_args))

        return

    def set_parameter(self, name, value):
        """
        Generates a BNG action command and adds it to the command queue

        Parameters
        ----------
        name: string
            The name of the parameter to set
        value: float-like
            Value of parameter

        """
        self.command_queue.write('\tsetParameter("%s", %g)\n' % (name, value))

    def set_concentration(self, cplx_pat, value):
        """
        Generates a BNG action command and adds it to the command queue

        Parameters
        ----------
        cplx_pat: pysb.ComplexPattern or string
            Species ComplexPattern, or a BNG format string representation
        value: float-like
            Initial concentration

        """
        if isinstance(cplx_pat, basestring):
            formatted_name = cplx_pat
        else:
            formatted_name = format_complexpattern(
                pysb.core.as_complex_pattern(cplx_pat)
            )
        self.command_queue.write('\tsetConcentration("%s", %g)\n' % (
            formatted_name, value))


def generate_hybrid_model(model, population_maps, additional_species=None,
                          safe=False, verbose=False, output_dir=None,
                          output_prefix=None, cleanup=True):
    with BngFileInterface(model,
                          output_dir=output_dir,
                          output_prefix=output_prefix,
                          cleanup=cleanup,
                          model_additional_species=additional_species,
                          model_population_maps=population_maps) as bng:
        bng.action('generate_hybrid_model', verbose=verbose, safe=safe,
                   suffix='hpp')
        bng.execute()

        with open(bng.base_filename + '_hpp.bngl', 'r') as f:
            return f.read()


def run_ssa(model, t_end=10, n_steps=100, param_values=None, output_dir=None,
            output_file_basename=None, cleanup=True, verbose=False,
            **additional_args):
    """
    Simulate a model with BNG's SSA simulator and return the trajectories.

    Parameters
    ----------
    model : Model
        Model to simulate.
    t_end : number, optional
        Final time point of the simulation.
    n_steps : int, optional
        Number of steps in the simulation.
    param_values : vector-like or dictionary, optional
            Values to use for every parameter in the model. Ordering is
            determined by the order of model.parameters. 
            If not specified, parameter values will be taken directly from
            model.parameters.
    output_dir : string, optional
        Location for temporary files generated by BNG. If None (the
        default), uses a temporary directory provided by the system. A
        temporary directory with a random name is created within the
        supplied location.
    output_file_basename : string, optional
        This argument is used as a prefix for the temporary BNG
        output directory, rather than the individual files.
    cleanup : bool, optional
        If True (default), delete the temporary files after the simulation is
        finished. If False, leave them in place. Useful for debugging.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.
    additional_args: kwargs, optional
        Additional arguments to pass to BioNetGen

    """
    bng_action_debug = verbose if isinstance(verbose, bool) else \
        verbose <= EXTENDED_DEBUG

    additional_args['method'] = 'ssa'
    additional_args['t_end'] = t_end
    additional_args['n_steps'] = n_steps
    additional_args['verbose'] = additional_args.get('verbose',
                                                     bng_action_debug)

    if param_values is not None:
        if len(param_values) != len(model.parameters):
            raise Exception("param_values must be the same length as "
                            "model.parameters")
        for i in range(len(param_values)):
            model.parameters[i].value = param_values[i]

    with BngFileInterface(model, verbose=verbose, output_dir=output_dir,
                          output_prefix=output_file_basename,
                          cleanup=cleanup) as bngfile:
        bngfile.action('generate_network', overwrite=True,
                       verbose=bng_action_debug)
        bngfile.action('simulate', **additional_args)

        bngfile.execute()

        output = bngfile.read_netfile()

        # Parse netfile (in case this hasn't already been done)
        if not model.species:
            _parse_netfile(model, iter(output.split('\n')))

        yfull = bngfile.read_simulation_results()

    return yfull


def generate_network(model, cleanup=True, append_stdout=False,
                     verbose=False, **kwargs):
    """
    Return the output from BNG's generate_network function given a model.

    The output is a BNGL model definition with additional sections 'reactions'
    and 'groups', and the 'species' section expanded to contain all possible
    species. BNG refers to this as a 'net' file.

    Parameters
    ----------
    model : Model
        Model to pass to generate_network.
    cleanup : bool, optional
        If True (default), delete the temporary files after the simulation is
        finished. If False, leave them in place (in `output_dir`). Useful for
        debugging.
    append_stdout : bool, optional
        This option is no longer supported and has been left here for API
        compatibility reasons.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.
    """
    bng_action_debug = verbose if isinstance(verbose, bool) else \
        verbose <= EXTENDED_DEBUG

    with BngFileInterface(model, verbose=verbose, cleanup=cleanup) as bngfile:
        bngfile._logger.info('Generating reaction network')
        bngfile.action('generate_network', overwrite=True,
                       verbose=bng_action_debug, **kwargs)
        bngfile.execute()

        output = bngfile.read_netfile()

    return output


def load_equations(model, netfile):
    """
    Load model equations from a specified netfile

    Useful for large models where BioNetGen network generation takes a long
    time - the .net file can be saved and reloaded using this function at a
    later date.

    Parameters
    ----------
    model: pysb.Model
        PySB model file
    netfile: str
        BNG netfile
    """
    if model.reactions:
        return
    with open(netfile, 'r') as f:
        _parse_netfile(model, f)


def generate_equations(model, cleanup=True, verbose=False, **kwargs):
    """
    Generate math expressions for reaction rates and species in a model.

    This fills in the following pieces of the model:

    * species
    * reactions
    * reactions_bidirectional
    * observables (just `coefficients` and `species` fields for each element)
    
    Parameters
    ----------
    model : Model
        Model to pass to generate_network.
    cleanup : bool, optional
        If True (default), delete the temporary files after the simulation is
        finished. If False, leave them in place (in `output_dir`). Useful for
        debugging.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.

    """
    # only need to do this once
    # TODO track "dirty" state, i.e. "has model been modified?"
    #   or, use a separate "math model" object to contain ODEs
    if model.reactions:
        return
    lines = iter(generate_network(model, cleanup=cleanup,
                                  verbose=verbose, **kwargs).split('\n'))
    _parse_netfile(model, lines)


def _parse_netfile(model, lines):
    """ Parse species, rxns and groups from a BNGL net file """
    try:
        while 'begin parameters' not in next(lines):
            pass
        while True:
            line = next(lines)
            if 'end parameters' in line: break
            _parse_parameter(model, line)

        while 'begin species' not in next(lines):
            pass
        model.species = []
        while True:
            line = next(lines)
            if 'end species' in line: break
            _parse_species(model, line)

        while 'begin reactions' not in next(lines):
            pass

        reaction_cache = {}
        while True:
            line = next(lines)
            if 'end reactions' in line: break
            _parse_reaction(model, line, reaction_cache=reaction_cache)
        # fix up reactions whose reverse version we saw first
        for r in model.reactions_bidirectional:
            if all(r['reverse']):
                r['reactants'], r['products'] = r['products'], r['reactants']
                r['rate'] *= -1
            # now the 'reverse' value is no longer needed
            del r['reverse']

        while 'begin groups' not in next(lines):
            pass
        while True:
            line = next(lines)
            if 'end groups' in line: break
            _parse_group(model, line)

    except StopIteration as e:
        pass


def _parse_parameter(model, line):
    _, pname, pval, _, ptype = line.strip().split()
    par_names = model.components.keys()
    if pname not in par_names:
        # Need to parse the value even for constants, since BNG considers some
        # expressions to be "Constant", e.g. "2^2"
        parsed_expr = parse_bngl_expr(pval)
        if ptype == 'Constant' and pname not in model._derived_parameters.keys():
            p = pysb.core.Parameter(pname, parsed_expr, _export=False)
            model._derived_parameters.add(p)
        elif ptype == 'ConstantExpression' and \
                pname not in model._derived_expressions.keys():
            p = pysb.core.Expression(pname, parsed_expr, _export=False)
            model._derived_expressions.add(p)
        else:
            raise ValueError('Unknown type {} for parameter {}'.format(
                ptype, pname))


def _parse_species(model, line):
    """Parse a 'species' line from a BNGL net file."""
    index, species, value = line.strip().split()
    species_compartment_name, complex_string = \
        re.match(r'(?:@(\w+)::)?(.*)', species).groups()
    species_compartment = model.compartments.get(species_compartment_name)
    monomer_strings = complex_string.split('.')
    monomer_patterns = []
    for ms in monomer_strings:
        monomer_name, site_strings, monomer_compartment_name = \
            re.match(r'\$?(\w+)\(([^)]*)\)(?:@(\w+))?', ms).groups()
        site_conditions = collections.defaultdict(list)
        if len(site_strings):
            for ss in site_strings.split(','):
                # FIXME this should probably be done with regular expressions
                if '!' in ss and '~' in ss:
                    site_name, condition = ss.split('~')
                    state, bond = condition.split('!')
                    if bond == '?':
                        bond = pysb.core.WILD
                    elif bond == '!':
                        bond = pysb.core.ANY
                    else:
                        bond = int(bond)
                    condition = (state, bond)
                elif '!' in ss:
                    site_name, condition = ss.split('!', 1)
                    if '!' in condition:
                        condition = [int(c) for c in condition.split('!')]
                    else:
                        condition = int(condition)
                elif '~' in ss:
                    site_name, condition = ss.split('~')
                else:
                    site_name, condition = ss, None
                site_conditions[site_name].append(condition)

        site_conditions = {k: v[0] if len(v) == 1 else pysb.core.MultiState(*v)
                           for k, v in site_conditions.items()}
        monomer = model.monomers[monomer_name]
        monomer_compartment = model.compartments.get(monomer_compartment_name)
        # Compartment prefix notation in BNGL means "assign this compartment to
        # all molecules without their own explicit compartment".
        compartment = monomer_compartment or species_compartment
        mp = pysb.core.MonomerPattern(monomer, site_conditions, compartment)
        monomer_patterns.append(mp)

    cp = pysb.core.ComplexPattern(monomer_patterns, None)
    model.species.append(cp)


def _parse_reaction(model, line, reaction_cache):
    """Parse a 'reaction' line from a BNGL net file."""
    (number, reactants, products, rate, rule) = line.strip().split(' ', 4)
    # the -1 is to switch from one-based to zero-based indexing
    reactants = tuple(int(r) - 1 for r in reactants.split(',') if r != '0')
    products = tuple(int(p) - 1 for p in products.split(',') if p != '0')
    rate = rate.rsplit('*')
    (rule_list, unit_conversion) = re.match(
        r'#([\w,\(\)]+)(?: unit_conversion=(.*))?\s*$',
        rule).groups()
    rule_list = rule_list.split(',')  # BNG lists all rules that generate a rxn
    # Support new (BNG 2.2.6-stable or greater) and old BNG naming convention
    # for reverse rules
    rule_name, is_reverse = zip(*[re.subn('^_reverse_|\(reverse\)$', '', r)
                                  for r in rule_list])
    is_reverse = tuple(bool(i) for i in is_reverse)
    r_names = ['__s%d' % r for r in reactants]
    rate_param = [model.parameters.get(r) or model.expressions.get(r) or
                  model._derived_parameters.get(r) or
                  model._derived_expressions.get(r) or float(r) for r in rate]
    combined_rate = sympy.Mul(*[sympy.S(t) for t in r_names + rate_param])
    reaction = {
        'reactants': reactants,
        'products': products,
        'rate': combined_rate,
        'rule': rule_name,
        'reverse': is_reverse,
        }
    model.reactions.append(reaction)
    # bidirectional reactions
    key = (reactants, products)
    key_reverse = (products, reactants)
    if key in reaction_cache:
        reaction_bd = reaction_cache.get(key)
        reaction_bd['rate'] += combined_rate
        reaction_bd['rule'] += tuple(r for r in rule_name if r not in
                                     reaction_bd['rule'])
    elif key_reverse in reaction_cache:
        reaction_bd = reaction_cache.get(key_reverse)
        reaction_bd['reversible'] = True
        reaction_bd['rate'] -= combined_rate
        reaction_bd['rule'] += tuple(r for r in rule_name if r not in
                                     reaction_bd['rule'])
    else:
        # make a copy of the reaction dict
        reaction_bd = dict(reaction)
        # default to false until we find a matching reverse reaction
        reaction_bd['reversible'] = False
        reaction_cache[key] = reaction_bd
        model.reactions_bidirectional.append(reaction_bd)


def _parse_group(model, line):
    """Parse a 'group' line from a BNGL net file."""
    # values are number (which we ignore), name, and species list
    values = line.strip().split()
    obs = model.observables[values[1]]
    if len(values) == 3:
        # combination is a comma separated list of [coeff*]speciesnumber
        for product in values[2].split(','):
            terms = product.split('*')
            # if no coeff given (just species), insert a coeff of 1
            if len(terms) == 1:
                terms.insert(0, 1)
            obs.coefficients.append(int(terms[0]))
            # -1 to change to 0-based indexing
            obs.species.append(int(terms[1]) - 1)


def _convert_tokens(tokens, local_dict, global_dict):
    for pos, tok in enumerate(tokens):
        if tok == (tokenize.NAME, 'if'):
            tokens[pos] = (tokenize.NAME, '__bngl_if')
            local_dict['__bngl_if'] = sympy.Function('bngl_if')
        elif tok == (tokenize.OP, '^'):
            tokens[pos] = (tokenize.OP, '**')
        elif tok == (tokenize.NAME, 'and'):
            tokens[pos] = (tokenize.OP, '&')
        elif tok == (tokenize.NAME, 'or'):
            tokens[pos] = (tokenize.OP, '|')
    return tokens


def parse_bngl_expr(text, *args, **kwargs):
    """Convert a BNGL math expression string to a sympy Expr."""

    # Translate a few operators with simple text replacement.
    text = text.replace('()', '')
    text = text.replace('==', '=')
    # Use sympy to parse the text into an Expr.
    trans = (
        sympy_parser.standard_transformations
        + (sympy_parser.convert_equals_signs, _convert_tokens)
    )
    expr = sympy_parser.parse_expr(text, *args, transformations=trans, **kwargs)
    # Transforming 'if' to Piecewise requires subexpression rearrangement, so we
    # use sympy's replace functionality rather than attempt it using text
    # replacements above.
    expr = expr.replace(
        sympy.Function('bngl_if'),
        lambda cond, t, f: sympy.Piecewise((t, cond), (f, True))
    )
    # Check for unsupported constructs.
    if expr.has('time'):
        raise ValueError(
            "Expressions referencing simulation time are not supported"
        )
    return expr


class NoInitialConditionsError(RuntimeError):
    """Model initial_conditions is empty."""
    def __init__(self):
        RuntimeError.__init__(self, "Model has no initial conditions or "
                                    "zero-order synthesis rules")


class NoRulesError(RuntimeError):
    """Model rules is empty."""
    def __init__(self):
        RuntimeError.__init__(self, "Model has no rules")



from abc import ABCMeta, abstractmethod
import numpy as np
import itertools
import sympy
import collections
import numbers
from pysb.core import MonomerPattern, ComplexPattern, as_complex_pattern, \
                      Parameter, Expression, Tag
from pysb.logging import get_logger, EXTENDED_DEBUG
import pickle
from pysb import __version__ as PYSB_VERSION
from datetime import datetime
import dateutil.parser
import copy
from warnings import warn
from pysb.pattern import SpeciesPatternMatcher
try:
    basestring
except NameError:
    # Python 3 compatibility.
    basestring = str

try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import h5py
except ImportError:
    h5py = None


class SimulatorException(Exception):
    pass


class Simulator(object):
    """An abstract base class for numerical simulation of models.

    .. warning::
        The interface for this class is considered experimental and may
        change without warning as PySB is updated.

    Parameters
    ----------
    model : pysb.Model
        Model to simulate.
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range. Returned trajectories are sampled at every value unless
        the simulation is interrupted for some reason, e.g., due to
        satisfaction
        of a logical stopping criterion (see 'tout' below).
    initials : vector-like or dict, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initials (with initial condition
        parameter values taken from `param_values` if specified).
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.

    Attributes
    ----------
    verbose: bool
        Verbosity flag passed to the constructor.
    model : pysb.Model
        Model passed to the constructor.
    tspan : vector-like
        Time values passed to the constructor.

    Notes
    -----
    If ``tspan`` is not defined, it may be defined in the call to the
    ``run`` method.

    The dimensionality of ``tout`` depends on whether a single simulation
    or multiple simulations are run.

    The dimensionalities of ``y``, ``yobs``, ``yobs_view``, ``yexpr``, and
    ``yexpr_view`` depend on the number of simulations being run as well
    as on the type of simulation, i.e., spatial vs. non-spatial.

    """
    __metaclass__ = ABCMeta

    _supports = { 'multi_initials' : False,
                  'multi_param_values' : False }

    @abstractmethod
    def __init__(self, model, tspan=None, initials=None,
                 param_values=None, verbose=False, **kwargs):
        # Get or create base a PySB logger for this module and model
        self._logger = get_logger(self.__module__, model=model,
                                  log_level=verbose)
        self._logger.debug('Simulator created')
        self._model = model
        self.verbose = verbose
        self.tout = None
        # Per-run initial conditions/parameter/tspan override
        self._tspan = tspan
        # Per-run tspan, initials and param_values
        self._run_tspan = None
        self._run_initials = None
        self._run_params = None
        # Base initials and param values
        self._params = None
        self.param_values = param_values
        self._initials = None
        self.initials = initials
        # Store init kwargs and run kwargs if needed for saving results
        self._init_kwargs = kwargs
        self._run_kwargs = None

    @property
    def model(self):
        return self._model

    @property
    def tspan(self):
        return self._run_tspan if self._run_tspan is not None else self._tspan

    @tspan.setter
    def tspan(self, new_tspan):
        self._tspan = new_tspan

    @staticmethod
    def _num_sims_calc(initials_or_params):
        """ Calculate number of simulations implied by initials or param
        values """
        if initials_or_params is None:
            return None

        if isinstance(initials_or_params, np.ndarray):
            return len(initials_or_params)

        first_entry = next(iter(initials_or_params.values()))

        try:
            return len(first_entry)  # First entry is iterable
        except TypeError:
            return 1  # First entry is non-iterable, e.g. int, float

    @property
    def initials_length(self):
        try:
            return len(self.initials)
        except SimulatorException:
            # Network free simulators
            if self._initials:
                return len(list(self._initials.values())[0])
            elif self._run_initials:
                return len(list(self._run_initials.values())[0])
            else:
                return len(self.param_values)

    def _update_initials_dict(self, initials_dict, initials_source, subs=None):
        if isinstance(initials_source, collections.Mapping):
            # Can't just use .update() as we need to test
            # equality with .is_equivalent_to()
            for cp, value_obj in initials_source.items():
                cp = as_complex_pattern(cp)
                if any(existing_cp.is_equivalent_to(cp)
                       for existing_cp in initials_dict):
                    continue

                if isinstance(value_obj, (collections.Sequence, np.ndarray))\
                        and all(isinstance(v, numbers.Number) for v in value_obj):
                    value = value_obj
                elif isinstance(value_obj, Expression):
                    value = [value_obj.expand_expr().evalf(subs=subs[sim]) for sim in range(len(subs))]
                elif isinstance(value_obj, Parameter):
                    # Set parameter using param_values
                    pi = self._model.parameters.index(value_obj)
                    value = [self.param_values[sim][pi] for sim in range(len(self.param_values))]
                else:
                    raise TypeError("Unexpected initial condition "
                                    "value type: %s" % type(value_obj))

                initials_dict[cp] = value
        elif initials_source is not None:
            # Update from array-like structure, which we can only do if we
            # have the species available (e.g. not in network-free simulations)
            if not self.model.species:
                raise ValueError(
                    'Cannot update initials from an array-like source without '
                    'model species.')
            for cp_idx, cp in enumerate(self.model.species):
                if any(existing_cp.is_equivalent_to(cp) for existing_cp in
                       initials_dict):
                    continue
                initials_dict[cp] = [initials_source[n][cp_idx]
                                     for n in range(len(initials_source))]
        return initials_dict

    @property
    def initials_dict(self):
        n_sims = self._check_run_initials_vs_base_initials_length()
        if n_sims == 1:
            n_sims = len(self.param_values)

        # Apply any per-run initial overrides
        initials_dict = self._update_initials_dict({}, self._run_initials)

        # Apply any base initial overrides
        initials_dict = self._update_initials_dict(initials_dict,
                                                   self._initials)

        model_initials = {ic.pattern: ic.value
                          for ic in self.model.initials}

        # Otherwise, populate initials from the model
        n_sims_params = len(self.param_values)
        n_sims_actual = max(n_sims_params, n_sims)

        # Get remaining initials from the model itself and
        # self.param_values, if necessary
        subs = None
        if any(isinstance(v, Expression) for v in model_initials.values()):
            # Only need parameter substitutions if model initials include
            # expressions
            subs = [
                dict((p, pv[i]) for i, p in
                     enumerate(self._model.parameters))
                for pv in self.param_values]
            if len(subs) == 1 and n_sims_actual > 1:
                subs = list(itertools.repeat(subs[0], n_sims_actual))

        initials_dict = self._update_initials_dict(
            initials_dict, model_initials, subs=subs
        )

        return initials_dict

    def _check_run_initials_vs_base_initials_length(self):
        # Otherwise, build the list from the model, and any overrides
        # specified in self._initials and self._run_initials
        n_sims_initials = self._num_sims_calc(self._initials)
        n_sims_run = self._num_sims_calc(self._run_initials)

        if n_sims_initials is not None and n_sims_run is not None \
                and n_sims_run != n_sims_initials:
            raise ValueError(
                "The base initials set with self.initials imply {} "
                "simulations, but the run() initials imply {} simulations."
                " Either set self.initials=None, or change the number of "
                "simulations in the run() initials".format(
                    n_sims_initials, n_sims_run))

        if n_sims_initials is not None:
            return n_sims_initials
        elif n_sims_run is not None:
            return n_sims_run
        else:
            return 1

    @property
    def initials(self):
        if not self.model.species:
            raise SimulatorException('No model species list - either '
                                     'generate the model equations or use '
                                     'initials_dict() for network-free '
                                     'simulations')

        # Check potential quick return options
        if self._run_initials is not None:
            if not isinstance(self._run_initials, collections.Mapping) and \
                    self._initials is None:
                return self._run_initials
        elif not isinstance(self._initials, collections.Mapping) and \
                self._initials is not None:
            return self._initials

        # At this point (after dimensionality check), we can return
        # self._run_initials if it's not a dictionary and not None
        if self._run_initials is not None and not isinstance(
                self._run_initials, collections.Mapping):
            return self._run_initials

        n_sims_initials = self._check_run_initials_vs_base_initials_length()
        n_sims_params = len(self.param_values)
        n_sims_actual = max(n_sims_params, n_sims_initials)

        y0 = np.full((n_sims_actual, len(self.model.species)), 0.0)

        for species, vals in self.initials_dict.items():
            species_index = self._model.get_species_index(species)
            y0[:, species_index] = vals

        return y0

    @initials.setter
    def initials(self, new_initials):
        self._initials = self._process_incoming_initials(new_initials)

    def _process_incoming_initials(self, new_initials):
        if new_initials is None:
            return None

        # If new_initials is a pandas dataframe, convert to a dict
        if pd and isinstance(new_initials, pd.DataFrame):
            new_initials = new_initials.to_dict(orient='list')

        # If new_initials is a list, convert to numpy array
        if isinstance(new_initials, list):
            new_initials = np.array(new_initials, copy=False)

        # Check if new_initials is a dict, and if so validate the keys
        # (ComplexPatterns)
        if isinstance(new_initials, dict):
            n_sims = 1
            if len(new_initials) > 0:
                n_sims = self._num_sims_calc(new_initials)
            for cplx_pat, val in new_initials.items():
                if not isinstance(cplx_pat, (MonomerPattern,
                                             ComplexPattern)):
                    raise ValueError('Dictionary key %s is not a '
                                     'MonomerPattern or ComplexPattern' %
                                     repr(cplx_pat))
                # if val is a number, convert it to a single-element array
                if not isinstance(val, (collections.Sequence, np.ndarray)):
                    val = [val]
                    new_initials[cplx_pat] = np.array(val)
                # otherwise, check whether simulator supports multiple
                # initial values :
                if len(val) != n_sims:
                    raise ValueError("all arrays in new_initials dictionary "
                                     "must be equal length")
                if not np.isfinite(val).all():
                    raise ValueError('Please check initial {} for non-finite '
                                     'values'.format(cplx_pat))
        elif isinstance(new_initials, np.ndarray):
            # if new_initials is a 1D array, convert to a 2D array of length 1
            if len(new_initials.shape) == 1:
                new_initials = np.resize(new_initials, (1, len(new_initials)))
            n_sims = new_initials.shape[0]
            # make sure number of initials values equals len(model.species)
            if new_initials.shape[1] != len(self._model.species):
                raise ValueError("new_initials must be the same length as "
                                 "model.species")
            if not np.isfinite(new_initials).all():
                raise ValueError('Please check initials array '
                                 'for non-finite values')
        else:
            raise ValueError(
                'Implicit conversion of data type "{}" is not '
                'supported. Please supply initials as a numpy array, list, '
                'or a pandas DataFrame.'.format(type(new_initials)))

        if n_sims > 1:
            if not self._supports['multi_initials']:
                raise ValueError(
                    self.__class__.__name__ +
                    " does not support multiple initial values at this time.")
            if 1 < len(self.param_values) != n_sims:
                raise ValueError(
                    'Cannot set initials for {} simulations '
                    'when param_values has been set for {} '
                    'simulations'.format(
                        n_sims, len(self.param_values)))

        return new_initials

    @property
    def param_values(self):
        if not self.model._derived_parameters:
            if self._params is not None and \
                    not isinstance(self._params, dict) and \
                    self._run_params is None:
                return self._params
            elif self._run_params is not None and \
                    not isinstance(self._run_params, dict) and \
                    self._params is None:
                return self._run_params

        # create parameter vector from the values in the model
        param_values_dict = {}
        n_sims = self._num_sims_calc(self._params)
        if isinstance(self._params, dict):
            param_values_dict.update(self._params)
        elif isinstance(self._params, np.ndarray):
            param_values_dict = dict(zip(
                [p.name for p in self._model.parameters], self._params.T))

        n_sims_run = self._num_sims_calc(self._run_params)

        if n_sims is None:
            n_sims = n_sims_run
        elif n_sims_run is not None and n_sims_run != n_sims:
            raise ValueError(
                "The base parameters set with self.param_values imply "
                "{} simulations, but the run() params imply {} "
                "simulations. Either set self.param_values=None, or "
                "change the number of simulations in the run() params"
                .format(n_sims, n_sims_run))

        # At this point (after dimensionality check) we can return the
        # _run_params, if it's not a dict
        if self._run_params is not None:
            if not isinstance(self._run_params, dict):
                if not self._model._derived_parameters:
                    return self._run_params
                else:
                    param_values_dict.update(dict(zip(
                        self.model.parameters.keys(), self._run_params
                    )))
            else:
                param_values_dict.update(self._run_params)

        if n_sims is None:
            n_sims = 1

        # Get the base parameters from the model
        param_values = np.array(
            [p.value for p in self._model.parameters] +
            [p.value for p in self._model._derived_parameters]
        )
        param_values = np.repeat([param_values], n_sims, axis=0)
        # Process overrides
        for key in param_values_dict.keys():
            try:
                pi = self._model.parameters.index(
                                self._model.parameters[key])
            except KeyError:
                raise IndexError("new_params dictionary has unknown "
                                 "parameter name (%s)" % key)
            # loop over n_sims
            for n in range(n_sims):
                param_values[n][pi] = param_values_dict[key][n]

        return param_values

    @param_values.setter
    def param_values(self, new_params):
        self._params = self._process_incoming_params(new_params)

    def _process_incoming_params(self, new_params):
        if new_params is None:
            return None

        # Convert pandas dataframe to dictionary
        if pd and isinstance(new_params, pd.DataFrame):
            new_params = new_params.to_dict(orient='list')

        # If new_params is a list, convert to numpy array
        if isinstance(new_params, list):
            new_params = np.array(new_params)

        if isinstance(new_params, dict):
            n_sims = 1
            if len(new_params) > 0:
                n_sims = self._num_sims_calc(new_params)
            for key, val in new_params.items():
                if key not in self._model.parameters.keys():
                    raise IndexError("new_params dictionary has unknown "
                                     "parameter name (%s)" % key)
                # if val is a number, convert it to a single-element array
                if not isinstance(val, collections.Sequence):
                    val = [val]
                    new_params[key] = np.array(val)
                # Check all elements are the same length
                if len(val) != n_sims:
                    raise ValueError("all arrays in params dictionary "
                                     "must be equal length")
        elif isinstance(new_params, np.ndarray):
            # if new_params is a 1D array, convert to a 2D array of length 1
            if len(new_params.shape) == 1:
                new_params = np.resize(new_params, (1, len(new_params)))
            n_sims = new_params.shape[0]
            # make sure number of param values equals len(model.parameters)
            if new_params.shape[1] != len(self._model.parameters):
                raise ValueError("new_params must be the same length as "
                                 "model.parameters")
        else:
            raise ValueError(
                'Implicit conversion of data type "{}" is not '
                'supported. Please supply parameters as a numpy array, list, '
                'or a pandas DataFrame.'.format(type(new_params)))

        # Check whether simulator supports multiple param_values
        if n_sims > 1 and not self._supports['multi_param_values']:
            raise ValueError(
                self.__class__.__name__ +
                " does not support multiple parameter values at this time.")
        return new_params

    def _reset_run_overrides(self):
        """
        Reset any single-run tspan, initials, param_values

        When calling run(), the user can specify tspan, initials and
        param_values, which are only used for a single run. This method
        resets those overrides after the run is complete (called from
        :func:`SimulationResult.__init__`).
        """
        self._run_tspan = None
        self._run_initials = None
        self._run_params = None

    @abstractmethod
    def run(self, tspan=None, initials=None, param_values=None,
            _run_kwargs=None):
        """Run a simulation.

        Notes for developers implementing Simulator subclasses:

        Implementations should return a :class:`.SimulationResult` object.
        Subclasses should pass any additional arguments run as a dictonary
        to the `_run_kwargs` argument when calling the superclass's run
        method. If the run method has variable keyword arguments, this can
        be achieved by passing `_run_kwargs=locals()` to the superclass's
        run method. The run kwargs are used for reference when saving and
        loading SimulationResults to disk. They aren't compulsory, but not
        including them will generate a warning. To suppress (e.g. if there
        are no additional arguments), set `_run_kwargs=[]`.
        """
        self._logger.info('Simulation(s) started')
        if _run_kwargs:
            # Don't store these arguments twice
            _run_kwargs.pop('self')
            _run_kwargs.pop('initials', None)
            _run_kwargs.pop('param_values', None)
            _run_kwargs.pop('tspan', None)
            self._run_kwargs = _run_kwargs
        elif _run_kwargs is None:
            self._logger.warning(
                '{} has not passed any additional run arguments to '
                '_run_kwargs. Instructions are included in the Simulation '
                'base class run method docstring.'.format(
                    self.__class__.__name__))
        self._run_tspan = tspan
        if self.tspan is None:
            raise ValueError("tspan must be defined before "
                             "simulation can run")
        self._run_params = self._process_incoming_params(param_values)
        self._run_initials = self._process_incoming_initials(initials)

        # If only one set of param_values, run all simulations
        # with the same parameters
        if len(self.param_values) == 1 and self.initials_length > 1:
            new_params = np.repeat(self.param_values,
                                   self.initials_length,
                                   axis=0)
            self._run_params = new_params

        # Error checks on 'param_values' and 'initials'
        if len(self.param_values) != self.initials_length:
            raise ValueError(
                    "'param_values' and 'initials' must be equal lengths.\n"
                    "len(param_values): %d\n"
                    "len(initials): %d" %
                    (len(self.param_values), self.initials_length))
        elif len(self.param_values.shape) != 2 or \
                self.param_values.shape[1] != (
                    len(self._model.parameters) +
                    len(self._model._derived_parameters)):
            raise ValueError(
                    "'param_values' must be a 2D array of dimension N_SIMS x "
                    "len(model.parameters).\n"
                    "param_values.shape: " + str(self.param_values.shape) +
                    "\nlen(model.parameters): %d" %
                    len(self._model.parameters))

        if self.model.species and (len(self.initials.shape) != 2 or
                self.initials.shape[1] != len(self._model.species)):
            raise ValueError(
                    "'initials' must be a 2D array of dimension N_SIMS x "
                    "len(model.species).\n"
                    "initials.shape: " + str(self.initials.shape) +
                    "\nlen(model.species): %d" % len(self._model.species))

        return None


class SimulationResult(object):
    """
    Results of a simulation with properties and methods to access them.

    .. warning::
        Please note that the interface for this class is considered
        experimental and may change without warning as PySB is updated.

    Notes
    -----
    In the attribute descriptions, a "trajectory set" is a 2D numpy array,
    species on first axis and time on second axis, with each element
    containing the concentration or count of the species at the specified time.

    A list of trajectory sets contains a trajectory set for each simulation.

    Parameters
    ----------
    simulator : Simulator
        The simulator object that generated the trajectories
    tout: list-like
        Time points returned by the simulator (may be different from ``tspan``
        if simulation is interrupted for some reason).
    trajectories : list or numpy.ndarray
        A set of species trajectories from a simulation. Should either be a
        list of 2D numpy arrays or a single 3D numpy array.
    squeeze : bool, optional (default: True)
        Return trajectories as a 2D array, rather than a 3d array, if only
        a single simulation was performed.
    simulations_per_param_set : int
        Number of trajectories per parameter set. Typically always 1 for
        deterministic simulators (e.g. ODE), but with stochastic simulators
        multiple trajectories per parameter/initial condition set are often
        desired.
    model: pysb.Model
    initials: numpy.ndarray
    param_values: numpy.ndarray
        model, initials, param_values are an alternative constructor
        mechanism used when loading SimulationResults from files (see
        :func:`SimulationResult.load`). Setting just the simulator argument
        instead of these arguments is recommended.

    Examples
    --------
    The following examples use a simple model with three observables and one
    expression, with a single simulation.

    >>> from pysb.examples.expression_observables import model
    >>> from pysb.simulator import ScipyOdeSimulator
    >>> import numpy as np
    >>> np.set_printoptions(precision=4)
    >>> sim = ScipyOdeSimulator(model, tspan=np.linspace(0, 40, 10), \
                                integrator_options={'atol': 1e-20})
    >>> simulation_result = sim.run()

    ``simulation_result`` is a :class:`SimulationResult` object. An
    observable can be accessed like so:

    >>> print(simulation_result.observables['Bax_c0']) \
        #doctest: +NORMALIZE_WHITESPACE
    [1.0000e+00   1.1744e-02   1.3791e-04   1.6196e-06   1.9020e-08
     2.2337e-10   2.6232e-12   3.0806e-14   3.6178e-16   4.2492e-18]

    It is also possible to retrieve the value of all observables at a
    particular time point, e.g. the final concentrations:

    >>> print(simulation_result.observables[-1]) \
        #doctest: +SKIP
    (4.2492e-18,   1.6996e-16,  1.)

    Expressions are read in the same way as observables:

    >>> print(simulation_result.expressions['NBD_signal']) \
        #doctest: +NORMALIZE_WHITESPACE
    [0.   4.7847  4.9956  4.9999  5.   5.   5.   5.   5.   5. ]

    The species trajectories can be accessed as a numpy ndarray:

    >>> print(simulation_result.species) #doctest: +NORMALIZE_WHITESPACE
    [[1.0000e+00   0.0000e+00   0.0000e+00]
     [1.1744e-02   5.2194e-02   9.3606e-01]
     [1.3791e-04   1.2259e-03   9.9864e-01]
     [1.6196e-06   2.1595e-05   9.9998e-01]
     [1.9020e-08   3.3814e-07   1.0000e+00]
     [2.2337e-10   4.9637e-09   1.0000e+00]
     [2.6232e-12   6.9951e-11   1.0000e+00]
     [3.0806e-14   9.5840e-13   1.0000e+00]
     [3.6178e-16   1.2863e-14   1.0000e+00]
     [4.2492e-18   1.6996e-16   1.0000e+00]]

    Species, observables and expressions can be combined into a single numpy
    ndarray and accessed similarly. Here, the initial concentrations of all
    these entities are examined:

    >>> print(simulation_result.all[0]) #doctest: +SKIP
    ( 1.,  0.,  0.,  1.,  0.,  0.,  0.)

    The ``all`` array can be accessed as a pandas DataFrame object,
    which allows for more convenient indexing and access to pandas advanced
    functionality, such as indexing and slicing. Here, the concentrations of
    the observable ``Bax_c0`` and the expression ``NBD_signal`` are read at
    time points between 5 and 15 seconds:

    >>> df = simulation_result.dataframe
    >>> print(df.loc[5:15, ['Bax_c0', 'NBD_signal']]) \
        #doctest: +NORMALIZE_WHITESPACE
                 Bax_c0  NBD_signal
    time
    8.888889   0.000138    4.995633
    13.333333  0.000002    4.999927
    """
    CUSTOM_ATTR_PREFIX = 'usrattr_'

    def __init__(self, simulator, tout, trajectories=None,
                 observables_and_expressions=None, squeeze=True,
                 simulations_per_param_set=1,
                 model=None, initials=None, param_values=None):
        if simulator:
            simulator._logger.debug('SimulationResult constructor started')
            self._param_values = simulator.param_values.copy()
            try:
                self._initials = simulator.initials.copy()
            except SimulatorException:
                # Network free simulations don't have initials list, only dict
                self._initials = simulator.initials_dict.copy()
            self._model = copy.deepcopy(simulator._model)
            self.simulator_class = simulator.__class__
            self.init_kwargs = copy.deepcopy(simulator._init_kwargs)
            self.run_kwargs = copy.deepcopy(simulator._run_kwargs)
        else:
            self._param_values = param_values
            self._initials = initials
            self._model = model
            self.simulator_class = None
            self.init_kwargs = {}
            self.run_kwargs = {}

        self.squeeze = squeeze
        self.tout = np.asarray(tout)
        self._yfull = None
        self.n_sims_per_parameter_set = simulations_per_param_set
        self.pysb_version = PYSB_VERSION
        self.timestamp = datetime.now()
        self.custom_attrs = {}

        if trajectories is None and observables_and_expressions is None:
            raise ValueError('Need to supply at least one of species '
                             'trajectories or observables_and_expressions')

        if trajectories is not None and len(trajectories) > 0:
            # Validate incoming trajectories
            if getattr(trajectories, 'ndim', None) == 3:
                # trajectories is a 3D array, create a list of 2D arrays
                # This is just a view and doesn't copy the data
                self._y = [tr for tr in trajectories]
            else:
                # Not a 3D array, check for a list of 2D arrays
                try:
                    if any(tr.ndim != 2 for tr in trajectories):
                        raise AttributeError
                except (AttributeError, TypeError):
                    raise ValueError("trajectories should be a 3D array or a "
                                     "list of 2D arrays")
                self._y = trajectories

            self._nsims = len(self._y)
            if len(self.tout) != self.nsims:
                raise ValueError("Simulator tout should be the same length as "
                                 "trajectories")
            for i in range(self.nsims):
                if len(self.tout[i]) != self._y[i].shape[0]:
                    raise ValueError("The number of time points in tout[{0}] "
                                     "should match the trajectories array for "
                                     "simulation {0}".format(i))
                if self._y[i].shape[1] != len(self._model.species):
                    raise ValueError("The number of species in trajectory {0} "
                                     "should match length of "
                                     "model.species".format(i))
        else:
            self._y = None

        # Calculate ``yobs`` and ``yexpr`` based on values of ``y``
        exprs = self._model.expressions_dynamic(include_local=False)
        expr_names = [expr.name for expr in exprs]
        model_obs = self._model.observables
        obs_names = list(model_obs.keys())
        param_names = list(p.name for p in self._model.parameters)

        if not _allow_unicode_recarray():
            for name_list, name_type in zip(
                    (expr_names, obs_names, param_names),
                    ('Expression', 'Observable', 'Parameter')):
                for i, name in enumerate(name_list):
                    try:
                        name_list[i] = name.encode('ascii')
                    except UnicodeEncodeError:
                        error_msg = 'Non-ASCII compatible ' + \
                                    '%s names not allowed' % name_type
                        raise ValueError(error_msg)

        yobs_dtype = (list(zip(obs_names, itertools.repeat(float)))
                      if obs_names else float)
        yexpr_dtype = (list(zip(expr_names, itertools.repeat(float)))
                       if expr_names else float)

        if observables_and_expressions:
            # Observables and expression values are used as supplied
            self._nsims = len(observables_and_expressions)
            self._yobs_view = [observables_and_expressions[n][:, 0:(len(
                self._model.observables))] for n in range(self.nsims)]
            self._yexpr_view = [observables_and_expressions[n][:, (len(
                self._model.observables)):] for n in range(self.nsims)]

            self._yobs = [self._yobs_view[n].reshape(
                len(tout[n]) * len(obs_names)).view(dtype=yobs_dtype) for n
                          in range(self.nsims)]
            self._yexpr = [self._yexpr_view[n].reshape(
                len(tout[n]) * len(expr_names)).view(dtype=yexpr_dtype) for n
                          in range(self.nsims)]
        else:
            self._yobs = [np.ndarray((len(self.tout[n]),),
                                     dtype=yobs_dtype) for n in range(self.nsims)]
            self._yobs_view = [self._yobs[n].view(float).
                               reshape(len(self._yobs[n]), -1) for n in range(
                self.nsims)]
            self._yexpr = [np.ndarray((len(self.tout[n]),),
                                      dtype=yexpr_dtype) for n in range(
                self.nsims)]
            self._yexpr_view = [self._yexpr[n].view(float).reshape(len(
                self._yexpr[n]), -1) for n in range(self.nsims)]

            # loop over simulations
            sym_names = obs_names + param_names
            expanded_exprs = [sympy.lambdify(sym_names, expr.expand_expr(),
                                             "numpy") for expr in exprs]
            for n in range(self.nsims):
                if simulator:
                    simulator._logger.log(EXTENDED_DEBUG,
                                          'Evaluating exprs/obs %d/%d'
                                          % (n + 1, self.nsims))

                # observables
                for i, obs in enumerate(model_obs):
                    self._yobs_view[n][:, i] = (
                        self._y[n][:, obs.species] * obs.coefficients).sum(axis=1)

                # expressions
                sym_dict = dict((k, self._yobs[n][k]) for k in obs_names)
                sym_dict.update(dict((p.name, self.param_values[
                    n // self.n_sims_per_parameter_set][i]) for i, p in
                                enumerate(self._model.parameters)))
                for i, expr in enumerate(exprs):
                    self._yexpr_view[n][:, i] = expanded_exprs[i](**sym_dict)

        if simulator:
            simulator._reset_run_overrides()
            simulator._logger.debug('SimulationResult constructor finished')

    def _squeeze_output(self, trajectories):
        """
        Reduces trajectories to a 2D matrix if only one simulation present

        Can be disabled by setting self.squeeze to False
        """
        if self.nsims == 1 and self.squeeze:
            return trajectories[0]
        else:
            return trajectories

    @property
    def nsims(self):
        """ The number of simulations in this SimulationResult """
        return self._nsims

    @property
    def all(self):
        """
        Aggregate species, observables, and expressions trajectories into
        a numpy.ndarray with record-style data-type for return to the user.
        """
        if self._yfull is None:
            if self._y is None:
                yfull_dtype = []
            else:
                sp_names = ['__s%d' % i
                            for i in range(len(self._model.species))]
                yfull_dtype = list(zip(sp_names, itertools.repeat(float)))
            if len(self._model.observables):
                yfull_dtype += self._yobs[0].dtype.descr
            if len(self._model.expressions_dynamic()):
                yfull_dtype += self._yexpr[0].dtype.descr
            yfull = []
            # loop over simulations
            for n in range(self.nsims):
                yfull.append(np.ndarray(len(self.tout[n]), yfull_dtype))
                yfull_view = yfull[n].view(float).reshape((len(yfull[n]), -1))
                n_sp = self._y[n].shape[1] if self._y else 0
                n_ob = self._yobs_view[n].shape[1]
                n_ex = self._yexpr_view[n].shape[1]
                if self._y:
                    yfull_view[:, :n_sp] = self._y[n]
                yfull_view[:, n_sp:n_sp + n_ob] = self._yobs_view[n]
                yfull_view[:, n_sp + n_ob:n_sp + n_ob + n_ex] = \
                    self._yexpr_view[n]
            self._yfull = yfull

        return self._squeeze_output(self._yfull)

    @property
    def dataframe(self):
        """
        A conversion of the trajectory sets (species, observables and
        expressions for all simulations) into a single
        :py:class:`pandas.DataFrame`.
        """
        if pd is None:
            raise Exception('Please "pip install pandas" for this feature')
        sim_ids = (np.repeat(range(self.nsims), [len(t) for t in self.tout]))
        times = np.concatenate(self.tout)
        if self.nsims == 1 and self.squeeze:
            idx = pd.Index(times, name='time')
        else:
            idx = pd.MultiIndex.from_tuples(list(zip(sim_ids, times)),
                                            names=['simulation', 'time'])
        simdata = self.all
        if not isinstance(simdata, np.ndarray):
            simdata = np.concatenate(simdata)
        return pd.DataFrame(simdata, index=idx)

    @property
    def species(self):
        """
        List of trajectory sets. The first dimension contains species.
        """
        if self._y is None:
            raise ValueError('No trajectories are available for network-free '
                             'simulations')
        return self._squeeze_output(self._y)

    @property
    def observables(self):
        """
        List of trajectory sets. The first dimension contains observables.
        """
        if not self._model.observables:
            raise ValueError('Model has no observables')
        return self._squeeze_output(self._yobs)

    def observable(self, pattern):
        """
        Calculate a pattern's trajectories without adding to model

        This method calculates an observable "on demand" using
        any supplied MonomerPattern or ComplexPattern against the simulation
        result, without re-running the simulation.

        Note that the monomers within the supplied pattern are reconciled
        with the SimulationResult's internal copy of the model by name. This
        method only works on simulations which calculate species
        trajectories (i.e. it will not work on network-free simulations).

        Raises a ValueError if the pattern does not match at least one species.

        Parameters
        ----------
        pattern: pysb.MonomerPattern or pysb.ComplexPattern
            An observable pattern to match

        Returns
        -------
        pandas.Series
            Series containing the simulation trajectories for the specified
            observable

        Examples
        --------

        >>> from pysb import ANY
        >>> from pysb.examples import earm_1_0
        >>> from pysb.simulator import ScipyOdeSimulator
        >>> simres = ScipyOdeSimulator(earm_1_0.model, tspan=range(5)).run()
        >>> m = earm_1_0.model.monomers

        Observable of bound Bid:

        >>> simres.observable(m.Bid(b=ANY))
        time
        0    0.000000e+00
        1    1.190933e-12
        2    2.768582e-11
        3    1.609716e-10
        4    5.320530e-10
        dtype: float64

        Observable of AMito bound to mCytoC:

        >>> simres.observable(m.AMito(b=1) % m.mCytoC(b=1))
        time
        0    0.000000e+00
        1    1.477319e-77
        2    1.669917e-71
        3    5.076939e-69
        4    1.157400e-66
        dtype: float64
        """

        # Adjust the supplied pattern's monomer objects to match the
        # simulationresult's internal model
        if isinstance(pattern, MonomerPattern):
            self._update_monomer_pattern(pattern)
        elif isinstance(pattern, ComplexPattern):
            for mp in pattern.monomer_patterns:
                self._update_monomer_pattern(mp)
        else:
            raise ValueError('The pattern must be a MonomerPattern or '
                             'ComplexPattern')

        if self._y is None:
            raise ValueError('On demand observables can only be calculated '
                             'on simulations with species trajectories')

        obs_matches = SpeciesPatternMatcher(self._model).match(
            pattern, index=True, counts=True)

        if not obs_matches:
            raise ValueError('No species match the supplied observable '
                             'pattern')

        return self.dataframe.iloc[:, list(obs_matches.keys())].multiply(
            list(obs_matches.values())).sum(axis=1)

    def _update_monomer_pattern(self, pattern):
        """ Update a pattern's monomer objects to use internal model

        Internal function for in-place update of a pattern to replace its
        monomers with those from SimulationResult's model, matching by name.

        Raises ValueError if no monomer with the specified name is in the
        model.
        """
        mon_name = pattern.monomer.name
        try:
            new_mon = self._model.monomers[mon_name]
        except KeyError:
            raise ValueError('There was no monomer called "{}" in the model '
                             '"{}" at the time of simulation'.format(
                                mon_name, self._model.name))
        pattern.monomer = new_mon

    @property
    def expressions(self):
        """
        List of trajectory sets. The first dimension contains expressions.
        """
        if not self._model.expressions_dynamic():
            raise ValueError('Model has no dynamic expressions')
        return self._squeeze_output(self._yexpr)

    @property
    def initials(self):
        return self._initials

    @property
    def param_values(self):
        return self._param_values

    def save(self, filename, dataset_name=None, group_name=None,
             append=False, include_obs_exprs=False):
        """
        Save a SimulationResult to a file (HDF5 format)

        HDF5 is a hierarchical, binary storage format well suited to storing
        matrix-like data. Our implementation requires the h5py package.

        Each SimulationResult is treated as an HDF5 dataset, stored within a
        group which is specific to a model. In this way, it is possible to save
        multiple SimulationResults for a specific model.

        A group is first created in the HDF file root (see group_name
        argument). Within that group, a dataset "_model" has a pickled
        version of the PySB model. SimulationResult are stored as groups
        within the model group.

        The file hierarchy under group_name/dataset_name/ then consists of
        the following HDF5 gzip compressed HDF5 datasets: trajectories,
        param_values, initials, tout, observables (optional) and expressions
        (optional); and the following attributes:
        simulator_class (pickled Class), simulator_kwargs (pickled dict),
        squeeze (bool), simulations_per_param_set (int), pysb_version (str),
        timestamp (ISO 8601 format).

        Custom attributes can be stored in the SimulationResult's
        `custom_attrs` dictionary. Keys should be strings, values can be any
        picklable object. When saved to HDF5, these custom attributes will
        be prefixed with ``usrattr_``.

        Parameters
        ----------
        filename: str
            Filename to which the data will be saved
        dataset_name: str or None
            Dataset name. If None, it will default to 'result'. If the
            dataset_name already exists within the group, a ValueError is
            raised.
        group_name: str or None
            Group name. If None, will default to the name of the model.
        append: bool
            If False, raise IOError if the specified file already exists. If
            True, append to existing file (or create if it doesn't exist).
        include_obs_exprs: bool
            Whether to save observables and expressions in the file or not.
            If they are not included, they can be recreated from the model
            and species trajectories when loaded back into PySB, but you may
            wish to include them for use with external software, or if you
            have complex expressions which take a long time to compute.

        """
        if h5py is None:
            raise Exception('Please install the h5py package for this feature')

        if self._y is None and not include_obs_exprs:
            warn('This SimulationResult has no trajectories - '
                 'you will need to set include_obs_exprs=True if '
                 'you wish to save observables and expressions')

        if group_name is None:
            group_name = self._model.name
        if dataset_name is None:
            dataset_name = 'result'

        # np.void maps to bytes in HDF5.
        enpickle = lambda obj: np.void(pickle.dumps(obj, -1))

        with h5py.File(filename, 'a' if append else 'w-') as hdf:
            # Get or create the group
            try:
                grp = hdf.create_group(group_name)
                grp.create_dataset('_model', data=enpickle(self._model))
            except ValueError:
                grp = hdf[group_name]
                model = pickle.loads(grp['_model'][()])
                if model.name != self._model.name:
                    raise ValueError('SimulationResult model has name "{}", '
                                     'but the model in HDF5 file group "{}" '
                                     'has name "{}"'.format(self._model.name,
                                                            group_name,
                                                            model.name))

            # Create the result dataset, which is actually a nested HDF group
            dset = grp.create_group(dataset_name)
            if self._y is not None:
                dset.create_dataset('trajectories', data=self._y,
                                    compression='gzip', shuffle=True)
            if include_obs_exprs:
                dset.create_dataset('observables', data=self._yobs_view,
                                    compression='gzip', shuffle=True)
                dset.create_dataset('expressions', data=self._yexpr_view,
                                    compression='gzip', shuffle=True)
            dset.create_dataset('param_values', data=self.param_values,
                                compression='gzip', shuffle=True)
            if isinstance(self.initials, np.ndarray):
                dset.create_dataset('initials', data=self.initials,
                                    compression='gzip', shuffle=True)
            else:
                dset.create_dataset('initials_dict', data=enpickle(
                    self.initials))
            dset.create_dataset('tout', data=self.tout,
                                compression='gzip')
            dset.attrs['simulator_class'] = enpickle(self.simulator_class)
            dset.attrs['init_kwargs'] = enpickle(self.init_kwargs)
            dset.attrs['run_kwargs'] = enpickle(self.run_kwargs)
            dset.attrs['squeeze'] = self.squeeze
            dset.attrs['simulations_per_param_set'] = \
                self.n_sims_per_parameter_set
            dset.attrs['pysb_version'] = self.pysb_version
            dset.attrs['timestamp'] = datetime.isoformat(
                self.timestamp)
            # This is the range of ints that can be natively encoded in HDF5.
            int_min = np.iinfo(np.int64).min
            int_max = np.iinfo(np.uint64).max
            for attr_name, attr_val in self.custom_attrs.items():
                # Pass HDF5-native values straight through, pickling others.
                if (not (isinstance(attr_val,
                                    (basestring, bytes, float, complex))
                         or (isinstance(attr_val, numbers.Integral)
                             and int_min <= attr_val <= int_max))):
                    attr_val = enpickle(attr_val)
                dset.attrs[self.CUSTOM_ATTR_PREFIX + attr_name] = attr_val

    @classmethod
    def load(cls, filename, dataset_name=None, group_name=None):
        """
        Load a SimulationResult from a file (HDF5 format)

        For a description of the file format see :func:`save`

        Parameters
        ----------
        filename: str
            Filename from which to load data
        dataset_name: str or None
            Dataset name. Can be left as None when the group specified only
            contains one dataset, which will then be selected. If None and
            more than one dataset is in the group, a ValueError is raised.
        group_name: str or None
            Group name. This is typically the name of the model. Can be left as
            None when the file only contains one group, which will then be
            selected. If None and more than group is in the file a
            ValueError is raised.

        Returns
        -------
        SimulationResult
            Set of trajectories and associated metadata loaded from the file
        """
        if h5py is None:
            raise Exception('Please "pip install h5py" for this feature')

        with h5py.File(filename, 'r') as hdf:
            if group_name is None:
                groups = hdf.keys()
                if len(groups) > 1:
                    raise ValueError("group_name must be specified when file "
                                     "contains more than one group. Options "
                                     "are: {}".format(str(groups)))
                group_name = next(iter(hdf))

            grp = hdf[group_name]

            if dataset_name is None:
                datasets = list(grp.keys())
                datasets.remove('_model')
                if len(datasets) > 1:
                    raise ValueError("dataset_name must be specified when "
                                     "group contains more than one dataset. "
                                     "Options are: {}".format(str(datasets)))
                dataset_name = datasets[0]

            dset = grp[dataset_name]

            obs_and_exprs = None

            if 'observables' in dset.keys():
                obs_and_exprs = list(dset['observables'])

            if 'expressions' in dset.keys():
                exprs = dset['expressions']
                if obs_and_exprs is None:
                    obs_and_exprs = list(exprs)
                else:
                    for i in range(len(obs_and_exprs)):
                        obs_and_exprs[i] = np.concatenate(
                            [obs_and_exprs[i], exprs[i]],
                            axis=1
                        )

            trajectories = None
            try:
                trajectories = dset['trajectories']
            except KeyError:
                pass

            try:
                initials = np.array(dset['initials'])
            except KeyError:
                initials = pickle.loads(dset['initials_dict'][()])

            simres = cls(
                simulator=None,
                model=pickle.loads(grp['_model'][()]),
                initials=initials,
                param_values=np.array(dset['param_values']),
                tout=np.array(dset['tout']),
                trajectories=trajectories,
                observables_and_expressions=obs_and_exprs,
                squeeze=dset.attrs['squeeze'],
                simulations_per_param_set=dset.attrs[
                    'simulations_per_param_set']
            )
            simres.pysb_version = dset.attrs['pysb_version']
            simres.timestamp = dateutil.parser.parse(
                dset.attrs['timestamp'])
            simres.simulator_class = pickle.loads(
                dset.attrs['simulator_class'])
            simres.init_kwargs = pickle.loads(dset.attrs['init_kwargs'])
            simres.run_kwargs = pickle.loads(dset.attrs['run_kwargs'])
            for attr_name in dset.attrs.keys():
                if attr_name.startswith(cls.CUSTOM_ATTR_PREFIX):
                    orig_name = attr_name[len(cls.CUSTOM_ATTR_PREFIX):]
                    attr_val = dset.attrs[attr_name]
                    # Restore objects that were pickled for storage.
                    if isinstance(attr_val, np.void):
                        attr_val = pickle.loads(attr_val)
                    simres.custom_attrs[orig_name] = attr_val
            return simres


def _allow_unicode_recarray():
    """Return True if numpy recarray can take unicode data type.

    In python 2, numpy doesn't allow unicode strings as names in arrays even
    if they are ascii encodeable. This function tests this directly.
    """
    try:
        np.ndarray((1,), dtype=[(u'X', float)])
    except TypeError:
        return False
    return True



import numpy
from pysb.simulator import ScipyOdeSimulator


class Solver(object):
    """An interface for numeric integration of models.

    Parameters
    ----------
    model : pysb.Model
        Model to integrate.
    tspan : vector-like
        Time values over which to integrate. The first and last values define
        the time range, and the returned trajectories will be sampled at every
        value.
    use_analytic_jacobian : boolean, optional
        Whether to provide the solver a Jacobian matrix derived analytically
        from the model ODEs. Defaults to False. If False, the integrator may
        approximate the Jacobian by finite-differences calculations when
        necessary (depending on the integrator and settings).
    integrator : string, optional (default: 'vode')
        Name of the integrator to use, taken from the list of integrators known
        to :py:class:`scipy.integrate.ode`.
    cleanup : bool, optional
        If True (default), delete the temporary files after the simulation is
        finished. If False, leave them in place. Useful for debugging.
    verbose : bool, optional (default: False)
        Verbose output 
    integrator_options
        Additional parameters for the integrator.

    Attributes
    ----------
    model : pysb.Model
        Model passed to the constructor
    tspan : vector-like
        Time values passed to the constructor.
    y : numpy.ndarray
        Species trajectories. Dimensionality is ``(len(tspan),
        len(model.species))``.
    yobs : numpy.ndarray with record-style data-type
        Observable trajectories. Length is ``len(tspan)`` and record names
        follow ``model.observables`` names.
    yobs_view : numpy.ndarray
        An array view (sharing the same data buffer) on ``yobs``.
        Dimensionality is ``(len(tspan), len(model.observables))``.
    yexpr : numpy.ndarray with record-style data-type
        Expression trajectories. Length is ``len(tspan)`` and record names
        follow ``model.expressions_dynamic()`` names.
    yexpr_view : numpy.ndarray
        An array view (sharing the same data buffer) on ``yexpr``.
        Dimensionality is ``(len(tspan), len(model.expressions_dynamic()))``.
    integrator : scipy.integrate.ode
        Integrator object.

    Notes
    -----
    The expensive step of generating the code for the right-hand side of the
    model's ODEs is performed during initialization. If you need to integrate
    the same model repeatedly with different parameters then you should build a
    single Solver object and then call its ``run`` method as needed.

    """
    def __init__(self, model, tspan, use_analytic_jacobian=False,
                 integrator='vode', cleanup=True,
                 verbose=False, **integrator_options):
        self._sim = ScipyOdeSimulator(model, verbose=verbose, tspan=tspan,
                                     use_analytic_jacobian=
                                     use_analytic_jacobian,
                                     integrator=integrator, cleanup=cleanup,
                                     **integrator_options)
        self.result = None
        self._yexpr_view = None
        self._yobs_view = None

    @property
    def _use_inline(self):
        return ScipyOdeSimulator._use_inline

    @_use_inline.setter
    def _use_inline(self, use_inline):
        ScipyOdeSimulator._use_inline = use_inline

    @property
    def y(self):
        return self.result.species if self.result is not None else None

    @property
    def yobs(self):
        return self.result.observables if self.result is not None else None

    @property
    def yobs_view(self):
        if self._yobs_view is None:
            self._yobs_view = self.yobs.view(float).reshape(len(self.yobs), -1)
        return self._yobs_view

    @property
    def yexpr(self):
        return self.result.expressions if self.result is not None else None

    @property
    def yexpr_view(self):
        if self._yexpr_view is None:
            self._yexpr_view = self.yexpr.view(float).reshape(len(self.yexpr),
                                                              -1)
        return self._yexpr_view

    def run(self, param_values=None, y0=None):
        """Perform an integration.

        Returns nothing; access the Solver object's ``y``, ``yobs``, or
        ``yobs_view`` attributes to retrieve the results.

        Parameters
        ----------
        param_values : vector-like or dictionary, optional
            Values to use for every parameter in the model. Ordering is
            determined by the order of model.parameters. 
            If passed as a dictionary, keys must be parameter names.
            If not specified, parameter values will be taken directly from
            model.parameters.
        y0 : vector-like, optional
            Values to use for the initial condition of all species. Ordering is
            determined by the order of model.species. If not specified, initial
            conditions will be taken from model.initials (with initial condition
            parameter values taken from `param_values` if specified).
        """
        self._yobs_view = None
        self._yexpr_view = None
        self.result = self._sim.run(param_values=param_values, initials=y0)


def odesolve(model, tspan, param_values=None, y0=None, integrator='vode',
             cleanup=True, verbose=False, **integrator_options):
    """Integrate a model's ODEs over a given timespan.

    This is a simple function-based interface to integrating (a.k.a. solving or
    simulating) a model. If you need to integrate a model repeatedly with
    different parameter values or initial conditions (as in parameter
    estimation), using the Solver class directly will provide much better
    performance.

    Parameters
    ----------
    model : pysb.Model
        Model to integrate.
    tspan : vector-like
        Time values over which to integrate. The first and last values define
        the time range, and the returned trajectories will be sampled at every
        value.
    param_values : vector-like, optional
        Values to use for every parameter in the model. Ordering is determined
        by the order of model.parameters. If not specified, parameter values
        will be taken directly from model.parameters.
    y0 : vector-like, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initials (with initial condition
        parameter values taken from `param_values` if specified).
    integrator : string, optional
        Name of the integrator to use, taken from the list of integrators known
        to :py:class:`scipy.integrate.ode`.
    cleanup : bool, optional
        Remove temporary files after completion if True. Set to False for
        debugging purposes.
    verbose : bool, optionsal
        Increase verbosity of simulator output.
    integrator_options :
        Additional parameters for the integrator.

    Returns
    -------
    yfull : record array
        The trajectories calculated by the integration. The first dimension is
        time and its length is identical to that of `tspan`. The second
        dimension is species/observables and its length is the sum of the
        lengths of model.species and model.observables. The dtype of the array
        specifies field names: '__s0', '__s1', etc. for the species and
        observable names for the observables. See Notes below for further
        explanation and caveats.

    Notes
    -----
    This function was the first implementation of integration support and
    accordingly it has a few warts:

    * It performs expensive code generation every time it is called.

    * The returned array, with its record-style data-type, allows convenient
      selection of individual columns by their field names, but does not permit
      slice ranges or indexing by integers for columns. If you only need access
      to your model's observables this is usually not a problem, but sometimes
      it's more convenient to have a "regular" array. See Examples below for
      code to do this.

    The actual integration code has since been moved to the Solver class and
    split up such that the code generation is only performed on
    initialization. The model may then be integrated repeatedly with different
    parameter values or initial conditions with much better
    performance. Additionally, Solver makes the species trajectories available
    as a simple array and only uses the record array for the observables where
    it makes sense.

    This function now simply serves as a wrapper for creating a Solver object,
    calling its ``run`` method, and building the record array to return.

    Examples
    --------
    Simulate a model and display the results for an observable:

    >>> from pysb.examples.robertson import model
    >>> from numpy import linspace
    >>> numpy.set_printoptions(precision=4)
    >>> yfull = odesolve(model, linspace(0, 40, 10))
    >>> print(yfull['A_total'])            #doctest: +NORMALIZE_WHITESPACE
    [1.      0.899   0.8506  0.8179  0.793   0.7728  0.7557  0.7408  0.7277
    0.7158]

    Obtain a view on a returned record array which uses an atomic data-type and
    integer indexing (note that the view's data buffer is shared with the
    original array so there is no extra memory cost):

    >>> yfull.shape == (10, )
    True
    >>> print(yfull.dtype)                 #doctest: +NORMALIZE_WHITESPACE
    [('__s0', '<f8'), ('__s1', '<f8'), ('__s2', '<f8'), ('A_total', '<f8'),
    ('B_total', '<f8'), ('C_total', '<f8')]
    >>> print(yfull[0:4, 1:3])             #doctest: +ELLIPSIS
    Traceback (most recent call last):
      ...
    IndexError: too many indices...
    >>> yarray = yfull.view(float).reshape(len(yfull), -1)
    >>> yarray.shape == (10, 6)
    True
    >>> print(yarray.dtype)
    float64
    >>> print(yarray[0:4, 1:3])            #doctest: +NORMALIZE_WHITESPACE
    [[0.0000e+00   0.0000e+00]
     [2.1672e-05   1.0093e-01]
     [1.6980e-05   1.4943e-01]
     [1.4502e-05   1.8209e-01]]

    """
    integrator_options['integrator'] = integrator
    sim = ScipyOdeSimulator(model, tspan=tspan, cleanup=cleanup,
                            verbose=verbose, **integrator_options)
    simres = sim.run(param_values=param_values, initials=y0)
    return simres.all



import abc
import copy
from pysb.logging import get_logger


class MergeStrategy(abc.ABC):
    """ Base class for building model merging strategies """
    def __init__(self, model1, model2):
        self._log = get_logger(__name__, model=model1)
        self.m1 = model1
        self.m2 = model2
        self.m1_mapping = None
        self.m2_mapping = None

    @abc.abstractmethod
    def merge(self):
        """ Merge model1 and model2 to return a new, combined model """
        pass

    def _rename(self, source_namespace, target_namespace, name):
        """ Propose a new name for a component in target_namespace """
        new_name = name
        i = 1
        while new_name in source_namespace or new_name in target_namespace:
            new_name = '{}__{}'.format(name, i)
            i += 1

        return new_name

    def _combine_models(self):
        m_out = copy.deepcopy(self.m1)
        m_out.name += '_merge_{}'.format(self.m2.name)
        target_namespace = set(m_out.components.keys())
        source_namespace = set(self.m2.components.keys())
        m2_copy = copy.deepcopy(self.m2)

        self.m1_mapping = dict(zip(source_namespace, source_namespace))
        self.m2_mapping = {}

        for c in m2_copy.components:
            if c.name in target_namespace:
                c_name = self._rename(source_namespace,
                                      target_namespace,
                                      c.name)
                self._log.debug('{} {} was renamed to {}'.format(
                    c.__class__.__name__, c.name, c_name
                ))
                c.rename(c_name)
            else:
                c_name = c.name
            target_namespace.add(c_name)
            self.m2_mapping[c.name] = c_name
            m_out.add_component(c)

        # Copy initial conditions
        m_out.initial_conditions += m2_copy.initial_conditions

        # Copy annotations
        m_out.annotations += m2_copy.annotations

        return m_out


class NoOverlapMerge(MergeStrategy):
    """ Naive strategy in which all components are assumed to be distinct """

    def merge(self):
        return self._combine_models()



from pysb.simulator.base import Simulator, SimulationResult, SimulatorException
from pysb.bng import BngFileInterface, load_equations, generate_hybrid_model
import numpy as np
import logging
from pysb.logging import EXTENDED_DEBUG
from pysb.core import as_complex_pattern, Parameter, \
    InvalidComplexPatternException
import collections
import os


class BngSimulator(Simulator):
    """ Simulate a model using BioNetGen """
    _supports = {
        'multi_initials':     True,
        'multi_param_values': True
    }
    _SIMULATOR_TYPES = ['ssa', 'nf', 'pla', 'ode']

    def __init__(self, model, tspan=None, initials=None, param_values=None,
                 cleanup=True, verbose=False):
        super(BngSimulator, self).__init__(model, tspan=tspan,
                                           initials=initials,
                                           param_values=param_values,
                                           verbose=verbose)
        self.cleanup = cleanup
        self._outdir = None

    def run(self, tspan=None, initials=None, param_values=None, n_runs=1,
            method='ssa', output_dir=None, output_file_basename=None,
            cleanup=None, population_maps=None, **additional_args):
        """
        Simulate a model using BioNetGen

        Parameters
        ----------
        tspan: vector-like
            time span of simulation
        initials: vector-like, optional
            initial conditions of model
        param_values : vector-like or dictionary, optional
            Values to use for every parameter in the model. Ordering is
            determined by the order of model.parameters.
            If not specified, parameter values will be taken directly from
            model.parameters.
        n_runs: int
            number of simulations to run
        method : str
            Type of simulation to run. Must be one of:

                * 'ssa' - Stochastic Simulation Algorithm (direct method with
                  propensity sorting)

                * 'nf' - Stochastic network free simulation with NFsim.
                  Performs Hybrid Particle/Population simulation if
                  population_maps argument is supplied

                * 'pla' - Partioned-leaping algorithm (variant of tau-leaping
                  algorithm)

                * 'ode' - ODE simulation (Sundials CVODE algorithm)

        output_dir : string, optional
            Location for temporary files generated by BNG. If None (the
            default), uses a temporary directory provided by the system. A
            temporary directory with a random name is created within the
            supplied location.
        output_file_basename : string, optional
            This argument is used as a prefix for the temporary BNG
            output directory, rather than the individual files.
        cleanup : bool, optional
            If True (default), delete the temporary files after the
            simulation is finished. If False, leave them in place (Useful for
            debugging). The default value, None, means to use the value
            specified in :py:func:`__init__`.
        population_maps: list of PopulationMap
            List of :py:class:`PopulationMap` objects for hybrid
            particle/population modeling. Only used when method='nf'.
        additional_args: kwargs, optional
            Additional arguments to pass to BioNetGen

        Examples
        --------

        Simulate a model using network free simulation (NFsim):

        >>> from pysb.examples import robertson
        >>> from pysb.simulator.bng import BngSimulator
        >>> model = robertson.model
        >>> sim = BngSimulator(model, tspan=np.linspace(0, 1))
        >>> x = sim.run(n_runs=1, method='nf')


        """
        super(BngSimulator, self).run(tspan=tspan,
                                      initials=initials,
                                      param_values=param_values,
                                      _run_kwargs=locals()
                                      )

        if cleanup is None:
            cleanup = self.cleanup

        if method not in self._SIMULATOR_TYPES:
            raise ValueError("Method must be one of " +
                             str(self._SIMULATOR_TYPES))

        if method != 'nf' and population_maps:
            raise ValueError('population_maps argument is only used when '
                             'method is "nf"')

        if method == 'nf':
            if population_maps is not None and (not isinstance(
                    population_maps, collections.Iterable) or
                    any(not isinstance(pm, PopulationMap) for pm in
                        population_maps)):
                raise ValueError('population_maps should be a list of '
                                 'PopulationMap objects')
            model_additional_species = self.initials_dict.keys()
        else:
            model_additional_species = None

        tspan_lin_spaced = np.allclose(
            self.tspan,
            np.linspace(self.tspan[0], self.tspan[-1], len(self.tspan))
        )
        if method == 'nf' and (not tspan_lin_spaced or self.tspan[0] != 0.0):
            raise SimulatorException('NFsim requires tspan to be linearly '
                                     'spaced starting at t=0')

        # BNG requires t_start even when supplying sample_times
        additional_args['t_start'] = self.tspan[0]
        if tspan_lin_spaced:
            # Just supply t_end and n_steps
            additional_args['n_steps'] = len(self.tspan) - 1
            additional_args['t_end'] = self.tspan[-1]
        else:
            additional_args['sample_times'] = self.tspan

        additional_args['method'] = method
        additional_args['print_functions'] = True
        verbose_bool = self._logger.logger.getEffectiveLevel() <= logging.DEBUG
        extended_debug = self._logger.logger.getEffectiveLevel() <= \
                         EXTENDED_DEBUG
        additional_args['verbose'] = extended_debug
        params_names = [g.name for g in self._model.parameters]

        n_param_sets = self.initials_length
        total_sims = n_runs * n_param_sets

        self._logger.info('Running %d BNG %s simulations' % (total_sims,
                                                             method))

        model_to_load = None
        hpp_bngl = None

        if population_maps:
            self._logger.debug('Generating hybrid particle-population model')
            hpp_bngl = generate_hybrid_model(
                self._model,
                population_maps,
                model_additional_species,
                verbose=extended_debug)
        else:
            model_to_load = self._model

        with BngFileInterface(model_to_load,
                              verbose=verbose_bool,
                              output_dir=output_dir,
                              output_prefix=output_file_basename,
                              cleanup=cleanup,
                              model_additional_species=model_additional_species
                              ) as bngfile:
            if hpp_bngl:
                hpp_bngl_filename = os.path.join(bngfile.base_directory,
                                                 'hpp_model.bngl')
                self._logger.debug('HPP BNGL:\n\n' + hpp_bngl)
                with open(hpp_bngl_filename, 'w') as f:
                    f.write(hpp_bngl)
            if method != 'nf':
                # TODO: Write existing netfile if already generated
                bngfile.action('generate_network', overwrite=True,
                               verbose=extended_debug)
            if output_file_basename is None:
                prefix = 'pysb'
            else:
                prefix = output_file_basename

            sim_prefix = 0
            for pset_idx in range(n_param_sets):
                for n in range(len(params_names)):
                    bngfile.set_parameter(params_names[n],
                                          self.param_values[pset_idx][n])
                for cp, values in self.initials_dict.items():
                    if population_maps:
                        for pm in population_maps:
                            if pm.complex_pattern.is_equivalent_to(cp):
                                cp = pm.counter_species
                                break
                    bngfile.set_concentration(cp, values[pset_idx])
                for sim_rpt in range(n_runs):
                    tmp = additional_args.copy()
                    tmp['prefix'] = '{}{}'.format(prefix, sim_prefix)
                    bngfile.action('simulate', **tmp)
                    bngfile.action('resetConcentrations')
                    sim_prefix += 1
            if hpp_bngl:
                bngfile.execute(reload_netfile=hpp_bngl_filename,
                                skip_file_actions=True)
            else:
                bngfile.execute()
            if method != 'nf':
                load_equations(self.model, bngfile.net_filename)
            list_of_yfull = \
                BngFileInterface.read_simulation_results_multi(
                [bngfile.base_filename + str(n) for n in range(total_sims)])

        tout = []
        species_out = []
        obs_exp_out = []
        for i in range(total_sims):
            yfull = list_of_yfull[i]
            yfull_view = yfull.view(float).reshape(len(yfull), -1)

            tout.append(yfull_view[:, 0])

            if method == 'nf':
                obs_exp_out.append(yfull_view[:, 1:])
            else:
                species_out.append(yfull_view[:,
                                   1:(len(self.model.species) + 1)])
                if len(self.model.observables) or len(self.model.expressions):
                    obs_exp_out.append(yfull_view[:,
                                        (len(self.model.species) + 1):
                                        (len(self.model.species) + 1) +
                                        len(self.model.observables) +
                                        len(self.model.expressions_dynamic())])

        return SimulationResult(self, tout=tout, trajectories=species_out,
                                observables_and_expressions=obs_exp_out,
                                simulations_per_param_set=n_runs)


class PopulationMap(object):
    """
    Population map for BioNetGen hybrid particle/population simulation

    For use with the :class:`BngSimulator`.

    References
    ----------

    Hogg et al. 2014:
    http://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1003544

    BioNetGen HPP documentation:
    http://bionetgen.org/index.php/Hybrid_particle-population_model_generator

    """
    def __init__(self, complex_pattern, lumping_rate, counter_species=None):
        try:
            self.complex_pattern = as_complex_pattern(complex_pattern)
        except InvalidComplexPatternException:
            raise ValueError('complex_pattern must be a ComplexPattern')

        if not isinstance(lumping_rate, Parameter):
            raise ValueError('lumping_rate must be a %s' % Parameter.__class__)

        self.lumping_rate = lumping_rate
        if counter_species is None:
            self.counter_species = None
        else:
            self.counter_species = str(counter_species)

    def __repr__(self):
        return 'PopulationMap({}, {}, {})'.format(
            self.complex_pattern,
            self.lumping_rate,
            'None' if self.counter_species is None else '{}'.format(
                self.counter_species)
        )



# from pysb.examples.earm_1_3 import model as earm
from pysb.examples.bax_pore import model as bax_pore
from pysb.importers.bngl import model_from_bngl
from pysb.tests.test_importers import _bngl_location
from pysb.merge import NoOverlapMerge
from pysb.export.pysb_flat import PysbFlatExporter

def test_merge_basic():
    # egfr = model_from_bngl(_bngl_location('egfr_net'))
    # print(egfr.monomers)
    # print(earm.monomers)
    # print(bax_pore.monomers)
    m_merge = NoOverlapMerge(bax_pore, bax_pore).merge()
    print(PysbFlatExporter(m_merge).export())
    raise

